"""Represents the current snapshot of the environment."""

import logging
from typing import Set, Dict, Optional, List

from pddl_plus_parser.models import GroundedPredicate, State, PDDLObject, Domain, ActionCall, PDDLFunction

from sam_learning.core.vocabulary_creator import VocabularyCreator


class EnvironmentSnapshot:
    """class representing the snapshot of the environment."""

    vocabulary_creator: VocabularyCreator
    previous_state_predicates: Set[GroundedPredicate]
    next_state_predicates: Set[GroundedPredicate]
    previous_state_functions: Dict[str, PDDLFunction]
    next_state_functions: Dict[str, PDDLFunction]
    partial_domain: Domain
    logger: logging.Logger

    def __init__(self, partial_domain: Domain):
        self.logger = logging.getLogger(__name__)
        self.vocabulary_creator = VocabularyCreator()
        self.previous_state_predicates = set()
        self.next_state_predicates = set()
        self.previous_state_functions = {}
        self.next_state_functions = {}
        self.partial_domain = partial_domain

    def _create_state_discrete_snapshot(self, state: State, relevant_objects: Dict[str, PDDLObject]) -> Set[GroundedPredicate]:
        """Creates a snapshot of the state predicates.

        :param state: the state to create a snapshot of.
        :param relevant_objects: the relevant objects in the observation.
        """
        self.logger.debug("Creating a snapshot of the state predicates.")
        positive_state_predicates, negative_state_predicates = set(), set()
        vocabulary = self.vocabulary_creator.create_vocabulary(domain=self.partial_domain, observed_objects=relevant_objects)

        for lifted_predicate_name, vocabulary_predicates in vocabulary.items():
            if lifted_predicate_name not in state.state_predicates:
                # updates all the grounded predicates where there of a lifted predicate that does not appear in the state.
                for grounded_predicate in vocabulary_predicates:
                    grounded_signature = {
                        param_name: relevant_objects[object_name].type for param_name, object_name in grounded_predicate.object_mapping.items()
                    }
                    negative_state_predicates.add(
                        GroundedPredicate(
                            name=grounded_predicate.name,
                            signature=grounded_signature,
                            object_mapping=grounded_predicate.object_mapping,
                            is_positive=False,
                        )
                    )

                continue

            for grounded_vocabulary_predicate in vocabulary_predicates:
                for predicate in state.state_predicates[lifted_predicate_name]:
                    predicate.is_positive = True  # TODO: this is a hack, fix it
                    if predicate.object_mapping == grounded_vocabulary_predicate.object_mapping:
                        positive_state_predicates.add(grounded_vocabulary_predicate)
                        break

            # Updates the grounded predicates of the predicates that appear in the state.
            for grounded_predicate in vocabulary_predicates.difference(positive_state_predicates):
                grounded_signature = {
                    param_name: relevant_objects[object_name].type for param_name, object_name in grounded_predicate.object_mapping.items()
                }
                negative_state_predicates.add(
                    GroundedPredicate(
                        name=grounded_predicate.name,
                        signature=grounded_signature,
                        object_mapping=grounded_predicate.object_mapping,
                        is_positive=False,
                    )
                )

        return positive_state_predicates.union(negative_state_predicates)

    def _create_state_numeric_snapshot(self, state: State, relevant_objects: Dict[str, PDDLObject]) -> Dict[str, PDDLFunction]:
        """Creating a snapshot of the state functions.

        :param state: the state to create a snapshot of.
        :param relevant_objects: the relevant objects of the action or all the possible objects in the observation.
        :return: the snapshot of the state functions.
        """
        self.logger.debug("Creating a snapshot of the state functions.")
        result = {}
        for function_str, pddl_function in state.state_fluents.items():
            if len(pddl_function.signature.keys()) == 0 or set(pddl_function.signature.keys()).issubset(relevant_objects):
                result[function_str] = pddl_function

        return result

    def create_propositional_state_snapshot(
        self, state: State, current_action: ActionCall, observation_objects: Dict[str, PDDLObject]
    ) -> Set[GroundedPredicate]:
        """Creates a propositional snapshot of the current state.

        :param state: the state of the environment.
        :param current_action: the current action that is to be executed.
        :param observation_objects: the objects in the observation.
        """
        self.logger.debug("Creating a snapshot of the environment.")
        parameters_including_consts = current_action.parameters + list(self.partial_domain.constants.keys())
        relevant_objects = {
            object_name: object_data for object_name, object_data in observation_objects.items() if object_name in parameters_including_consts
        }
        return self._create_state_discrete_snapshot(state, relevant_objects)

    def create_numeric_state_snapshot(
        self, state: State, current_action: ActionCall, observation_objects: Dict[str, PDDLObject]
    ) -> Dict[str, PDDLFunction]:
        """Creates a numeric snapshot of the current state.

        :param state: the state of the environment.
        :param current_action: the current action that is to be executed.
        :param observation_objects: the objects in the observation.
        """
        self.logger.debug("Creating a snapshot of the environment.")
        parameters_including_consts = current_action.parameters + list(self.partial_domain.constants.keys())
        relevant_objects = {
            object_name: object_data for object_name, object_data in observation_objects.items() if object_name in parameters_including_consts
        }
        return self._create_state_numeric_snapshot(state, relevant_objects)

    def create_triplet_snapshot(
        self,
        previous_state: State,
        next_state: State,
        current_action: ActionCall,
        observation_objects: Dict[str, PDDLObject],
        specific_types: Optional[List[str]] = [],
    ) -> None:
        """Creates a snapshot of the environment.

        :param previous_state: the previous state of the environment.
        :param next_state: the next state of the environment.
        :param current_action: the current action.
        :param observation_objects: the objects in the observation.
        :param specific_types: the types of the objects to include in the snapshot.
        """
        self.logger.debug("Creating a snapshot of the environment.")
        parameters_including_consts = current_action.parameters + list(self.partial_domain.constants.keys())
        relevant_objects = {
            object_name: object_data
            for object_name, object_data in {**observation_objects, **self.partial_domain.constants}.items()
            if object_name in parameters_including_consts or object_data.type.name in specific_types
        }

        self.previous_state_predicates = self._create_state_discrete_snapshot(previous_state, relevant_objects)
        self.next_state_predicates = self._create_state_discrete_snapshot(next_state, relevant_objects)
        self.previous_state_functions = self._create_state_numeric_snapshot(previous_state, relevant_objects)
        self.next_state_functions = self._create_state_numeric_snapshot(next_state, relevant_objects)



import random

import numpy as np
import pytest
from pandas import DataFrame
from pddl_plus_parser.models import PDDLFunction, ObjectType

from sam_learning.core import NotSafeActionError
from sam_learning.core.numeric_learning.linear_regression_learner import LinearRegressionLearner
from sam_learning.core.numeric_learning.numeric_utils import create_monomials, create_polynomial_string

TEST_ACTION_NAME = "test_action"


@pytest.fixture
def linear_regression_learner() -> LinearRegressionLearner:
    domain_functions = {
        "x": PDDLFunction(name="x", signature={}),
        "y": PDDLFunction(name="y", signature={}),
        "z": PDDLFunction(name="z", signature={}),
        "w": PDDLFunction(name="w", signature={}),
    }
    return LinearRegressionLearner(TEST_ACTION_NAME, domain_functions)


@pytest.fixture
def polynomial_regression_learner() -> LinearRegressionLearner:
    domain_functions = {
        "(x )": PDDLFunction(name="x", signature={}),
        "(y )": PDDLFunction(name="y", signature={}),
        "(z )": PDDLFunction(name="z", signature={}),
        "(w )": PDDLFunction(name="w", signature={}),
    }
    return LinearRegressionLearner(TEST_ACTION_NAME, domain_functions, polynom_degree=1)


@pytest.fixture
def regression_learner_with_functions_with_the_same_name() -> LinearRegressionLearner:
    domain_functions = {
        "(x ?f2)": PDDLFunction(name="x", signature={"?f2": ObjectType}),
        "(x ?f1)": PDDLFunction(name="x", signature={"?f1": ObjectType}),
        "(cost )": PDDLFunction(name="cost", signature={}),
    }
    return LinearRegressionLearner(TEST_ACTION_NAME, domain_functions)


def test_add_new_observation_to_the_previous_state_data_adds_the_data_to_this_dataframe_only(linear_regression_learner: LinearRegressionLearner,):
    pre_state_data = {"(x )": 2.0, "(y )": 3.31, "(z )": -1.12, "(w )": 0.0}
    linear_regression_learner.add_new_observation(pre_state_data, store_in_prev_state=True)
    assert len(linear_regression_learner.previous_state_data) == 1
    assert len(linear_regression_learner.next_state_data) == 0


def test_add_new_observation_to_the_next_state_data_adds_the_data_to_this_dataframe_only(linear_regression_learner: LinearRegressionLearner,):
    next_state_data = {"(x )": 2.0, "(y )": 3.31, "(z )": -1.12, "(w )": 0.0}
    linear_regression_learner.add_new_observation(next_state_data, store_in_prev_state=False)
    assert len(linear_regression_learner.previous_state_data) == 0
    assert len(linear_regression_learner.next_state_data) == 1


def test_add_new_observation_with_missing_functions_in_the_data_removes_the_columns_from_the_dataframe(
    linear_regression_learner: LinearRegressionLearner,
):
    assert len(linear_regression_learner.next_state_data.columns.tolist()) == 4
    next_state_data = {"(x )": 2.0, "(y )": 3.31, "(z )": -1.12}
    linear_regression_learner.add_new_observation(next_state_data, store_in_prev_state=False)
    assert len(linear_regression_learner.previous_state_data.columns.tolist()) == 4
    assert len(linear_regression_learner.next_state_data.columns.tolist()) == 3


def test_add_new_observation_when_the_input_is_a_polynom_with_variable_multiplications_adds_all_permutations_to_previous_state_db_and_only_values_of_variable_to_post_state_db(
    linear_regression_learner: LinearRegressionLearner,
):
    pre_state_matrix = {
        "(x )": [2],
        "(y )": [0],
        "(z )": [1],
    }
    next_state_matrix = {
        "(x )": [2],
        "(y )": [2],
        "(z )": [0],
    }
    function_names = ["(x )", "(y )", "(z )"]
    linear_regression_learner.add_new_observation({name: pre_state_matrix[name][0] for name in function_names}, store_in_prev_state=True)
    linear_regression_learner.add_new_observation({name: next_state_matrix[name][0] for name in function_names}, store_in_prev_state=False)
    numeric_effects, preconditions, learned_correctly = linear_regression_learner.construct_assignment_equations()
    assert not learned_correctly
    assert preconditions is None
    assert len(numeric_effects) == 2
    for effect in numeric_effects:
        print(effect.to_pddl())


def test_extract_const_conditions_extract_preconditions_that_appear_in_all_samples_as_constant_a_single_time(
    linear_regression_learner: LinearRegressionLearner,
):
    precondition_statements = [
        ["(> (x) 0)", "(> (y) 0)", "(> (z) 0)"],
        ["(> (x) 0)", "(> (y) 1)", "(> (z) 2)"],
        ["(> (x) 0)", "(> (y) 2)", "(> (z) 4)"],
    ]
    const_conditions = linear_regression_learner._extract_const_conditions(precondition_statements)
    assert len(const_conditions) == 1
    assert const_conditions[0] == "(> (x) 0)"
    for precond in precondition_statements:
        assert len(precond) == 2


def test_validate_legal_equations_returns_true_when_the_number_of_equations_is_valid_and_the_matrix_has_one_solution(
    linear_regression_learner: LinearRegressionLearner,
):
    pre_state_data = {"x": [2, 1, 3, 0], "y": [3, -1, 2, 0], "z": [-1, 2, 4, 1]}
    dataframe = DataFrame(pre_state_data)
    assert linear_regression_learner._validate_legal_equations(dataframe)


def test_validate_legal_returns_false_when_there_are_not_enough_independent_equations(linear_regression_learner: LinearRegressionLearner,):
    pre_state_data = {"x": [0, 1, 2, 4], "y": [0, 2, 4, 8], "z": [0, 0, 18, 6]}
    dataframe = DataFrame(pre_state_data)
    assert not linear_regression_learner._validate_legal_equations(dataframe)


def test_validate_legal_equations_returns_false_when_the_number_of_equations_is_too_small(linear_regression_learner: LinearRegressionLearner,):
    pre_state_data = {"x": [2, 1, 3], "y": [3, -1, 2], "z": [-1, 2, 4]}
    dataframe = DataFrame(pre_state_data)
    assert not linear_regression_learner._validate_legal_equations(dataframe)


def test_solve_function_linear_equations_returns_correct_solution_for_a_solvable_matrix(linear_regression_learner: LinearRegressionLearner,):
    equation_matrix = {"x": [2, 1, 3, 0], "y": [3, -1, 2, 0], "z": [-1, 2, 4, 1], "label": [4, -1, 9, 15 / 35]}
    dataframe = DataFrame(equation_matrix)
    regression_array = np.array(dataframe.loc[:, dataframe.columns != "label"])
    function_post_values = np.array(dataframe["label"])
    try:
        coefficients, learning_score = linear_regression_learner._solve_regression_problem(regression_array, function_post_values)
        assert learning_score == 1
        assert len(coefficients) == 4

    except NotSafeActionError:
        pytest.fail()


def test_solve_function_linear_equations_raises_error_when_there_is_no_solution_with_the_correct_score_required(
    linear_regression_learner: LinearRegressionLearner,
):
    equation_matrix = {"x": [0, 1, 1, 0], "y": [0, 0, 0, 0], "z": [-1, 0, -1, 1], "label": [4, -1, 9, 15 / 35]}
    dataframe = DataFrame(equation_matrix)
    regression_array = np.array(dataframe.loc[:, dataframe.columns != "label"])
    function_post_values = np.array(dataframe["label"])
    with pytest.raises(NotSafeActionError) as e:
        linear_regression_learner._solve_regression_problem(regression_array, function_post_values, allow_unsafe_learning=False)


def test_compute_non_constant_change_returns_correct_pddl_equation_form_for_variable_with_assignment_sign(
    linear_regression_learner: LinearRegressionLearner,
):
    equation_matrix = {"x": [2, 1, 3, 0], "y": [3, -1, 2, 0], "z": [-1, 2, 4, 1], "label": [4, -1, 9, 15 / 35]}
    dataframe = DataFrame(equation_matrix)
    try:
        equation = linear_regression_learner._solve_linear_equations("w", dataframe)
        assert equation is not None
        assert equation.startswith("(assign")

    except NotSafeActionError:
        pytest.fail()


def test_compute_non_constant_change_returns_correct_pddl_equation_form_for_variable_with_circular_assignment_statement(
    linear_regression_learner: LinearRegressionLearner,
):
    equation_matrix = {"x": [2, 1, 3, 0], "y": [3, -1, 2, 0], "z": [-1, 2, 4, 1], "label": [4, -1, 9, 15 / 35]}
    dataframe = DataFrame(equation_matrix)
    try:
        equation = linear_regression_learner._solve_linear_equations("x", dataframe)
        assert equation is not None
        assert equation.startswith("(increase")

    except NotSafeActionError:
        pytest.fail()


def test_action_not_affects_fluent_returns_true_when_action_does_not_affect_fluent(linear_regression_learner: LinearRegressionLearner,):
    equation_matrix = {
        "(x)": [2, 3, 4, 5, 6, 7, 8, 9],
        "label": [2, 3, 4, 5, 6, 7, 8, 9],
    }
    dataframe = DataFrame(equation_matrix)
    result = linear_regression_learner._solve_linear_equations("(x)", dataframe)
    assert result is None


def test_action_not_affects_fluent_returns_false_when_action_does_affect_fluent(linear_regression_learner: LinearRegressionLearner,):
    equation_matrix = {
        "(x)": [2, 4, 6],
        "label": [4, 6, 8],
    }
    dataframe = DataFrame(equation_matrix)
    result = linear_regression_learner._solve_linear_equations("(x)", dataframe)
    assert result is not None
    assert result == "(increase (x) 2)"


def test_combine_states_data_creates_correct_dataframe_with_correct_values(linear_regression_learner: LinearRegressionLearner,):
    pre_state_matrix = {
        "(x )": [2, 1, 3, 0, 1],
        "(y )": [0, 3, 12, 43, 1],
        "(z )": [0, 2, 4, 1, 100],
    }
    next_state_matrix = {
        "(x )": [2, 6, 19, 44, 102],
        "(y )": [3, 6, 15, 46, 4],
        "(z )": [1, 1, 1, 1, 1],
    }
    function_names = ["(x )", "(y )", "(z )"]
    for i in range(5):
        linear_regression_learner.add_new_observation({name: pre_state_matrix[name][i] for name in function_names}, store_in_prev_state=True)
        linear_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    combined_state = linear_regression_learner._combine_states_data()
    assert len(combined_state.columns) == 6


def test_solve_safe_independent_equations_returns_correct_pddl_equation_form_for_variable_with_assignment_sign(
    linear_regression_learner: LinearRegressionLearner,
):
    equation_matrix = {"(x)": [2, 1, 3, 0], "(y)": [3, -1, 2, 0], "(z)": [-1, 2, 4, 1], "label": [21, 20, 22, 19]}
    dataframe = DataFrame(equation_matrix)
    constant_change = linear_regression_learner._solve_linear_equations("(x)", dataframe)
    assert constant_change is not None
    assert constant_change == "(increase (x) 19)"


def test_solve_safe_independent_equations_equation_form_when_linear_equations_are_solvable(linear_regression_learner: LinearRegressionLearner,):
    equation_matrix = {"(x)": [2, 1, 3, 0], "(y)": [3, -1, 2, 0], "(z)": [-1, 2, 4, 1], "label": [21, 23, 252, 129]}
    dataframe = DataFrame(equation_matrix)
    non_constant_change = linear_regression_learner._solve_linear_equations("(x)", dataframe)
    assert non_constant_change is not None


def test_construct_assignment_equations_returns_correct_equations_for_all_fluents(linear_regression_learner: LinearRegressionLearner,):
    pre_state_matrix = {
        "(x )": [2, 1, 3, 0, 1],
        "(y )": [0, 3, 12, 43, 1],
        "(z )": [0, 2, 4, 1, 100],
    }
    next_state_matrix = {"(x )": [2, 6, 19, 44, 102], "(y )": [3, 6, 15, 46, 4], "(z )": [1, 1, 1, 1, 1]}
    function_names = ["(x )", "(y )", "(z )"]
    for i in range(5):
        linear_regression_learner.add_new_observation({name: pre_state_matrix[name][i] for name in function_names}, store_in_prev_state=True)
        linear_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = linear_regression_learner.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 3
    assert preconditions is None
    for effect in numeric_effects:
        print(effect.to_pddl())


def test_construct_assignment_equations_returns_correct_equations_when_no_change_happens(linear_regression_learner: LinearRegressionLearner,):
    pre_state_matrix = {
        "(x )": [2, 1, 3, 0, 1],
        "(y )": [0, 3, 12, 43, 1],
        "(z )": [0, 2, 4, 1, 100],
    }
    next_state_matrix = {
        "(x )": [2, 1, 3, 0, 1],
        "(y )": [0, 3, 12, 43, 1],
        "(z )": [0, 2, 4, 1, 100],
    }
    function_names = ["(x )", "(y )", "(z )"]
    for i in range(5):
        linear_regression_learner.add_new_observation({name: pre_state_matrix[name][i] for name in function_names}, store_in_prev_state=True)
        linear_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = linear_regression_learner.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 0
    assert preconditions is None


def test_construct_assignment_equations_restrictive_preconditions_when_not_enough_data_is_given_and_the_values_are_equal_before_and_after_results_in_no_effect(
    linear_regression_learner: LinearRegressionLearner,
):
    pre_state_matrix = {
        "(x )": [2, 1],
        "(y )": [0, 3],
        "(z )": [0, 2],
    }
    next_state_matrix = {
        "(x )": [2, 1],
        "(y )": [0, 3],
        "(z )": [0, 2],
    }
    function_names = ["(x )", "(y )", "(z )"]
    for i in range(2):
        linear_regression_learner.add_new_observation({name: pre_state_matrix[name][i] for name in function_names}, store_in_prev_state=True)
        linear_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = linear_regression_learner.construct_assignment_equations()
    assert not learned_correctly
    assert preconditions is None  # The preconditions will be learned in externally to maintain safety.
    assert len(numeric_effects) == 0
    for effect in numeric_effects:
        print(effect.to_pddl())


def test_construct_assignment_equations_restrictive_preconditions_when_not_enough_data_is_given_and_the_values_are_not_equal_before_and_after_results_in_effect(
    linear_regression_learner: LinearRegressionLearner,
):
    pre_state_matrix = {
        "(x )": [2, 1],
        "(y )": [0, 3],
        "(z )": [0, 2],
    }
    next_state_matrix = {
        "(x )": [12, 3],
        "(y )": [1, 3],
        "(z )": [9, 2],
    }
    function_names = ["(x )", "(y )", "(z )"]
    for i in range(2):
        linear_regression_learner.add_new_observation({name: pre_state_matrix[name][i] for name in function_names}, store_in_prev_state=True)
        linear_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = linear_regression_learner.construct_assignment_equations()
    assert not learned_correctly
    assert preconditions is None  # The preconditions will be learned in externally to maintain safety.
    assert len(numeric_effects) == 3
    for effect in numeric_effects:
        print(effect.to_pddl())


def test_construct_assignment_equations_restrictive_preconditions_given_only_one_sample(polynomial_regression_learner: LinearRegressionLearner,):
    pre_state_matrix = {
        "(x )": [2],
        "(y )": [0],
        "(z )": [1],
    }
    next_state_matrix = {
        "(x )": [2],
        "(y )": [2],
        "(z )": [0],
    }
    function_names = ["(x )", "(y )", "(z )"]
    monomials = create_monomials(function_names, polynom_degree=1)
    polynomial_regression_learner.add_new_observation(
        {create_polynomial_string(monomial): np.prod([pre_state_matrix[name][0] for name in monomial]) for monomial in monomials},
        store_in_prev_state=True,
    )
    polynomial_regression_learner.add_new_observation({name: next_state_matrix[name][0] for name in function_names}, store_in_prev_state=False)
    assert len(polynomial_regression_learner.previous_state_data) == 1
    assert len(polynomial_regression_learner.next_state_data) == 1
    assert len(polynomial_regression_learner.previous_state_data.columns) == 6
    assert len(polynomial_regression_learner.next_state_data.columns) == 3


def test_construct_assignment_equations_with_polynomial_degree_one_returns_correct_equations_when_polynom_is_linear(
    polynomial_regression_learner: LinearRegressionLearner,
):
    numbers = list(range(101))
    x = random.choices(numbers, k=20)
    y = random.choices(numbers, k=20)
    z = random.choices(numbers, k=20)
    pre_state_matrix = {
        "(x )": x,
        "(y )": y,
        "(z )": z,
    }
    next_state_matrix = {
        "(x )": [2 * x1 + 3 * y1 + 4 * z1 + 5 for x1, y1, z1 in zip(x, y, z)],
        "(y )": y,
        "(z )": z,
    }
    function_names = ["(x )", "(y )", "(z )"]
    monomials = create_monomials(function_names, polynom_degree=1)
    for i in range(20):
        polynomial_regression_learner.add_new_observation(
            {create_polynomial_string(monomial): np.prod([pre_state_matrix[name][i] for name in monomial]) for monomial in monomials},
            store_in_prev_state=True,
        )
        polynomial_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = polynomial_regression_learner.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 1
    effect = numeric_effects.pop()
    assert effect.to_pddl() == "(increase (x ) (+ (x ) (+ (* (y ) 3) (+ (* (z ) 4) 5))))"


def test_construct_assignment_equations_with_polynomial_degree_one_returns_correct_equations_when_polynom_degree_is_also_one(
    polynomial_regression_learner: LinearRegressionLearner,
):
    k = 500
    numbers = list(range(1000))
    x = random.sample(numbers, k=k)
    y = random.sample(numbers, k=k)
    z = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x )": x,
        "(y )": y,
        "(z )": z,
    }
    next_state_matrix = {
        "(x )": [2 * x1 + 3 * y1 * x1 + 4 * z1 + 5 for x1, y1, z1 in zip(x, y, z)],
        "(y )": y,
        "(z )": z,
    }
    function_names = ["(x )", "(y )", "(z )"]
    monomials = create_monomials(function_names, polynom_degree=1)
    for i in range(k):
        polynomial_regression_learner.add_new_observation(
            {create_polynomial_string(monomial): np.prod([pre_state_matrix[name][i] for name in monomial]) for monomial in monomials},
            store_in_prev_state=True,
        )
        polynomial_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = polynomial_regression_learner.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 1
    effect = numeric_effects.pop()
    assert effect.to_pddl() == "(increase (x ) (+ (x ) (+ (* (z ) 4) (+ (* (* (x ) (y )) 3) 5))))"


def test_construct_assignment_equations_with_polynomial_degree_one_returns_correct_equations_when_polynom_degree_is_also_one_and_calculating_multiple_different_equations(
    polynomial_regression_learner: LinearRegressionLearner,
):
    k = 500
    numbers = list(range(1000))
    x = random.sample(numbers, k=k)
    y = random.sample(numbers, k=k)
    z = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x )": x,
        "(y )": y,
        "(z )": z,
    }
    next_state_matrix = {
        "(x )": [2 * x1 + 3 * y1 * x1 + 4 * z1 + 5 for x1, y1, z1 in zip(x, y, z)],
        "(y )": [y1 + y1 * x1 + 1 for x1, y1, in zip(x, y)],
        "(z )": [8 * x1 + 3 * y1 + 4 * z1 + 19 for x1, y1, z1 in zip(x, y, z)],
    }
    function_names = ["(x )", "(y )", "(z )"]
    monomials = create_monomials(function_names, polynom_degree=1)
    for i in range(k):
        polynomial_regression_learner.add_new_observation(
            {create_polynomial_string(monomial): np.prod([pre_state_matrix[name][i] for name in monomial]) for monomial in monomials},
            store_in_prev_state=True,
        )
        polynomial_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = polynomial_regression_learner.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 3
    expected_effects = {
        "(increase (x ) (+ (x ) (+ (* (z ) 4) (+ (* (* (x ) (y )) 3) 5))))",
        "(increase (z ) (+ (* (x ) 8) (+ (* (y ) 3) (+ (* (z ) 3) 19))))",
        "(increase (y ) (+ (* (x ) (y )) 1))",
    }
    for effect in numeric_effects:
        assert effect.to_pddl() in expected_effects


def test_construct_assignment_equations_with_polynomial_degree_one_returns_correct_equations_when_polynom_degree_is_also_one_and_there_is_and_assignment_equation(
    polynomial_regression_learner: LinearRegressionLearner,
):
    k = 500
    numbers = list(range(1000))
    x = random.sample(numbers, k=k)
    y = random.sample(numbers, k=k)
    z = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x )": x,
        "(y )": y,
        "(z )": z,
    }
    next_state_matrix = {
        "(x )": [2 * x1 + 3 * y1 * x1 + 4 * z1 + 5 for x1, y1, z1 in zip(x, y, z)],
        "(y )": [y1 + y1 * x1 + 1 for x1, y1, in zip(x, y)],
        "(z )": [1] * k,
    }
    function_names = ["(x )", "(y )", "(z )"]
    monomials = create_monomials(function_names, polynom_degree=1)
    for i in range(k):
        polynomial_regression_learner.add_new_observation(
            {create_polynomial_string(monomial): np.prod([pre_state_matrix[name][i] for name in monomial]) for monomial in monomials},
            store_in_prev_state=True,
        )
        polynomial_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = polynomial_regression_learner.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 3
    expected_effects = {
        "(increase (x ) (+ (x ) (+ (* (z ) 4) (+ (* (* (x ) (y )) 3) 5))))",
        "(assign (z ) 1)",
        "(increase (y ) (+ (* (x ) (y )) 1))",
    }
    for effect in numeric_effects:
        assert effect.to_pddl() in expected_effects


def test_construct_assignment_equations_with_polynomial_degree_one_and_relevant_fluents_does_not_fail_and_returns_correct_equations(
    polynomial_regression_learner: LinearRegressionLearner,
):
    k = 500
    numbers = list(range(1000))
    x = random.sample(numbers, k=k)
    y = random.sample(numbers, k=k)
    z = random.sample(numbers, k=k)
    w = random.sample(numbers, k=k)
    q = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x )": x,
        "(y )": y,
        "(z )": z,
        "(w )": w,
        "(q )": q,
    }
    next_state_matrix = {
        "(x )": [2 * x1 + 3 * y1 * x1 + 4 * z1 + 5 for x1, y1, z1 in zip(x, y, z)],
        "(y )": [y1 + y1 * x1 + 1 for x1, y1, in zip(x, y)],
        "(z )": [1] * k,
        "(w )": [7 * x1 + 82 * w1 + 1 * z1 + 5 for x1, w1, z1 in zip(x, w, z)],
        "(q )": [2 * w1 + 245 * y1 + 21 * q1 + 5 for w1, y1, q1 in zip(w, y, q)],
    }
    function_names = ["(x )", "(y )", "(z )", "(w )", "(q )"]
    monomials = create_monomials(function_names, polynom_degree=1)
    for i in range(k):
        polynomial_regression_learner.add_new_observation(
            {create_polynomial_string(monomial): np.prod([pre_state_matrix[name][i] for name in monomial]) for monomial in monomials},
            store_in_prev_state=True,
        )
        polynomial_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = polynomial_regression_learner.construct_assignment_equations(
        relevant_fluents=["(x )", "(* (x ) (y ))", "(y )", "(z )"]
    )
    assert learned_correctly
    assert len(numeric_effects) == 3
    expected_effects = {
        "(increase (x ) (+ (x ) (+ (* (z ) 4) (+ (* (* (x ) (y )) 3) 5))))",
        "(assign (z ) 1)",
        "(increase (y ) (+ (* (x ) (y )) 1))",
    }
    for effect in numeric_effects:
        assert effect.to_pddl() in expected_effects


def test_construct_assignment_equations_with_polynomial_degree_one_with_change_that_increase_by_constant_and_a_variable_calculates_result_correctly(
    polynomial_regression_learner: LinearRegressionLearner,
):
    k = 50
    numbers = list(range(500))
    x = random.sample(numbers, k=k)
    y = random.sample(numbers, k=k)
    z = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x )": x,
        "(y )": y,
        "(z )": z,
    }
    next_state_matrix = {
        "(x )": [x1 + 1 for x1 in x],
        "(y )": [y1 + x1 + 1 for x1, y1, in zip(x, y)],
        "(z )": [1] * k,
    }
    function_names = ["(x )", "(y )", "(z )"]
    monomials = create_monomials(function_names, polynom_degree=1)
    for i in range(k):
        polynomial_regression_learner.add_new_observation(
            {create_polynomial_string(monomial): np.prod([pre_state_matrix[name][i] for name in monomial]) for monomial in monomials},
            store_in_prev_state=True,
        )
        polynomial_regression_learner.add_new_observation({name: next_state_matrix[name][i] for name in function_names}, store_in_prev_state=False)

    numeric_effects, preconditions, learned_correctly = polynomial_regression_learner.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 3
    expected_effects = {
        "(increase (x ) 1)",
        "(assign (z ) 1)",
        "(increase (y ) (+ (x ) 1))",
    }
    for effect in numeric_effects:
        assert effect.to_pddl() in expected_effects


def test_construct_assignment_equations_when_two_functions_share_the_same_name_but_with_different_parameters_will_result_in_correct_assignment_of_the_equations_for_both_versions(
    regression_learner_with_functions_with_the_same_name: LinearRegressionLearner,
):
    k = 50
    numbers = list(range(500))
    x_f1 = random.sample(numbers, k=k)
    x_f2 = random.sample(numbers, k=k)
    cost = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x ?f1)": x_f1,
        "(x ?f2)": x_f2,
        "(cost )": cost,
    }
    next_state_matrix = {
        "(x ?f1)": [x1 - 4 for x1 in x_f1],
        "(x ?f2)": [x2 + 2 for x2 in x_f2],
        "(cost )": [c + 1 for c in cost],
    }
    for i in range(k):
        regression_learner_with_functions_with_the_same_name.add_new_observation(
            state_data={name: pre_state_matrix[name][i] for name in pre_state_matrix}, store_in_prev_state=True
        )
        regression_learner_with_functions_with_the_same_name.add_new_observation(
            {name: next_state_matrix[name][i] for name in next_state_matrix}, store_in_prev_state=False
        )

    numeric_effects, preconditions, learned_correctly = regression_learner_with_functions_with_the_same_name.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 3
    expected_effects = {
        "(decrease (x ?f1) 4)",
        "(increase (x ?f2) 2)",
        "(increase (cost ) 1)",
    }
    for effect in numeric_effects:
        assert effect.to_pddl() in expected_effects


def test_construct_assignment_equations_when_two_functions_share_the_same_name_but_with_different_parameters_will_result_in_correct_assignment_of_the_equations_for_both_versions_when_observing_only_a_few_samples(
    regression_learner_with_functions_with_the_same_name: LinearRegressionLearner,
):
    k = 3
    numbers = list(range(500))
    x_f1 = random.sample(numbers, k=k)
    cost = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x ?f1)": x_f1,
        "(x ?f2)": [0] * 3,
        "(cost )": cost,
    }
    next_state_matrix = {
        "(x ?f1)": [x1 - 4 for x1 in x_f1],
        "(x ?f2)": [2] * 3,
        "(cost )": [c + 1 for c in cost],
    }
    for i in range(k):
        regression_learner_with_functions_with_the_same_name.add_new_observation(
            state_data={name: pre_state_matrix[name][i] for name in pre_state_matrix}, store_in_prev_state=True
        )
        regression_learner_with_functions_with_the_same_name.add_new_observation(
            {name: next_state_matrix[name][i] for name in next_state_matrix}, store_in_prev_state=False
        )

    numeric_effects, preconditions, learned_correctly = regression_learner_with_functions_with_the_same_name.construct_assignment_equations()
    assert learned_correctly
    assert len(numeric_effects) == 3
    expected_effects = {
        "(decrease (x ?f1) 4)",
        "(assign (x ?f2) 2)",
        "(increase (cost ) 1)",
    }
    for effect in numeric_effects:
        assert effect.to_pddl() in expected_effects


def test_construct_assignment_equations_when_two_functions_share_the_same_name_but_with_different_parameters_identifies_that_action_does_not_affect_function_even_with_small_number_of_samples(
    regression_learner_with_functions_with_the_same_name: LinearRegressionLearner,
):
    k = 2
    numbers = list(range(500))
    x_f1 = random.sample(numbers, k=k)
    cost = random.sample(numbers, k=k)
    pre_state_matrix = {
        "(x ?f1)": x_f1,
        "(x ?f2)": [0] * 3,
        "(cost )": cost,
    }
    next_state_matrix = {
        "(x ?f1)": [x1 - 4 for x1 in x_f1],
        "(x ?f2)": [0] * 3,
        "(cost )": [c + 1 for c in cost],
    }
    for i in range(k):
        regression_learner_with_functions_with_the_same_name.add_new_observation(
            state_data={name: pre_state_matrix[name][i] for name in pre_state_matrix}, store_in_prev_state=True
        )
        regression_learner_with_functions_with_the_same_name.add_new_observation(
            {name: next_state_matrix[name][i] for name in next_state_matrix}, store_in_prev_state=False
        )

    numeric_effects, preconditions, learned_correctly = regression_learner_with_functions_with_the_same_name.construct_assignment_equations()
    assert len(numeric_effects) == 2



"""module tests for the Numeric SAM learning algorithm"""
import json
from typing import Dict, List

from pddl_plus_parser.lisp_parsers import ProblemParser, TrajectoryParser, DomainParser
from pddl_plus_parser.models import Domain, Problem, Observation, Predicate
from pytest import fixture

from sam_learning.learners.numeric_sam import NumericSAMLearner
from tests.consts import (
    DEPOT_FLUENTS_MAP_PATH,
    SATELLITE_FLUENTS_MAP_PATH,
    SATELLITE_PROBLEMATIC_PROBLEM_PATH,
    SATELLITE_PROBLEMATIC_NUMERIC_TRAJECTORY_PATH,
    MINECRAFT_FLUENTS_MAP_PATH,
    sync_snapshot,
    MINECRAFT_MEDIUM_DOMAIN_PATH,
    MINECRAFT_MEDIUM_TRAJECTORY_PATH,
    MINECRAFT_MEDIUM_FLUENTS_MAP_PATH,
    MINECRAFT_SMALL_DOMAIN_PATH,
    MINECRAFT_SMALL_TRAJECTORY_PATH,
    DRIVERLOG_POLY_DOMAIN_PATH,
    DRIVERLOG_POLY_PROBLEM_PATH,
    DRIVERLOG_POLY_TRAJECTORY_PATH,
    TEST_PPO_OBSERVATIONS_DIRECTORY,
    TEST_PPO_MINECRAFT_DOMAIN,
    FARMLAND_DOMAIN_PATH,
    FARMLAND_TRAJECTORIES_DIRECTORY,
)


@fixture()
def satellite_problem_problematic(satellite_numeric_domain: Domain) -> Problem:
    return ProblemParser(problem_path=SATELLITE_PROBLEMATIC_PROBLEM_PATH, domain=satellite_numeric_domain).parse_problem()


@fixture()
def satellite_observation_problematic(satellite_numeric_domain: Domain, satellite_problem_problematic: Problem) -> Observation:
    return TrajectoryParser(satellite_numeric_domain, satellite_problem_problematic).parse_trajectory(SATELLITE_PROBLEMATIC_NUMERIC_TRAJECTORY_PATH)


@fixture()
def minecraft_medium_preconditions_fluents_map() -> Dict[str, List[str]]:
    return json.load(open(MINECRAFT_MEDIUM_FLUENTS_MAP_PATH, "rt"))


@fixture()
def minecraft_medium_domain() -> Domain:
    return DomainParser(MINECRAFT_MEDIUM_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def minecraft_ppo_domain() -> Domain:
    return DomainParser(TEST_PPO_MINECRAFT_DOMAIN, partial_parsing=True).parse_domain()


@fixture()
def driverlog_polynomial_domain() -> Domain:
    return DomainParser(DRIVERLOG_POLY_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def driverlog_polynomial_problem(driverlog_polynomial_domain: Domain) -> Problem:
    return ProblemParser(problem_path=DRIVERLOG_POLY_PROBLEM_PATH, domain=driverlog_polynomial_domain).parse_problem()


@fixture()
def driverlog_polynomial_observation(driverlog_polynomial_domain: Domain, driverlog_polynomial_problem: Problem) -> Observation:
    return TrajectoryParser(driverlog_polynomial_domain, driverlog_polynomial_problem).parse_trajectory(DRIVERLOG_POLY_TRAJECTORY_PATH)


@fixture()
def driverlog_polynomial_nsam(driverlog_polynomial_domain: Domain) -> NumericSAMLearner:
    return NumericSAMLearner(driverlog_polynomial_domain, {action: [] for action in driverlog_polynomial_domain.actions}, polynomial_degree=1)


@fixture()
def minecraft_medium_observation(minecraft_medium_domain: Domain) -> Observation:
    return TrajectoryParser(minecraft_medium_domain).parse_trajectory(MINECRAFT_MEDIUM_TRAJECTORY_PATH)


@fixture()
def minecraft_medium_sam(minecraft_medium_domain: Domain, minecraft_medium_preconditions_fluents_map: Dict[str, List[str]]) -> NumericSAMLearner:
    return NumericSAMLearner(minecraft_medium_domain, minecraft_medium_preconditions_fluents_map)


@fixture()
def minecraft_small_domain() -> Domain:
    return DomainParser(MINECRAFT_SMALL_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def minecraft_small_observation(minecraft_small_domain: Domain) -> Observation:
    return TrajectoryParser(minecraft_small_domain).parse_trajectory(MINECRAFT_SMALL_TRAJECTORY_PATH)


@fixture()
def minecraft_small_sam(minecraft_small_domain: Domain) -> NumericSAMLearner:
    return NumericSAMLearner(minecraft_small_domain)


@fixture()
def depot_fluents_map() -> Dict[str, List[str]]:
    with open(DEPOT_FLUENTS_MAP_PATH, "rt") as json_file:
        return json.load(json_file)


@fixture()
def minecraft_fluents_map() -> Dict[str, List[str]]:
    with open(MINECRAFT_FLUENTS_MAP_PATH, "rt") as json_file:
        return json.load(json_file)


@fixture()
def satellite_fluents_map() -> Dict[str, List[str]]:
    with open(SATELLITE_FLUENTS_MAP_PATH, "rt") as json_file:
        return json.load(json_file)


@fixture()
def depot_nsam(depot_domain: Domain, depot_fluents_map: Dict[str, List[str]]) -> NumericSAMLearner:
    return NumericSAMLearner(depot_domain, depot_fluents_map)


@fixture()
def satellite_nsam(satellite_numeric_domain: Domain, satellite_fluents_map: Dict[str, List[str]]) -> NumericSAMLearner:
    return NumericSAMLearner(satellite_numeric_domain, satellite_fluents_map)


@fixture()
def minecraft_nsam(minecraft_domain: Domain, minecraft_fluents_map: Dict[str, List[str]]) -> NumericSAMLearner:
    return NumericSAMLearner(minecraft_domain, minecraft_fluents_map)


@fixture()
def farmland_domain() -> Domain:
    return DomainParser(FARMLAND_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def farmland_nsam(farmland_domain: Domain) -> NumericSAMLearner:
    nsam = NumericSAMLearner(farmland_domain, polynomial_degree=0)
    return nsam


@fixture()
def elevators_nsam(elevators_domain: Domain) -> NumericSAMLearner:
    nsam = NumericSAMLearner(elevators_domain, polynomial_degree=0)
    return nsam


def test_add_new_action_adds_action_to_fluents_storage(depot_nsam: NumericSAMLearner, depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    assert action_call.name in depot_nsam.storage


def test_add_new_action_adds_discrete_preconditions_to_the_learned_action(depot_nsam: NumericSAMLearner, depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if isinstance(precondition, Predicate)]
    assert len(action_discrete_preconditions) > 0


def test_add_new_action_adds_the_required_predicates_to_the_action_preconditions(depot_nsam: NumericSAMLearner, depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if isinstance(precondition, Predicate)]
    preconditions_str = {precondition.untyped_representation for precondition in action_discrete_preconditions}
    assert preconditions_str.issuperset({"(at ?x ?y)"})


def test_update_action_updates_action_in_the_storage(depot_nsam: NumericSAMLearner, depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    depot_nsam.update_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    assert action_call.name in depot_nsam.storage


def test_update_action_adds_discrete_preconditions_to_the_learned_action(depot_nsam: NumericSAMLearner, depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    depot_nsam.update_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if isinstance(precondition, Predicate)]
    assert len(action_discrete_preconditions) > 0


def test_update_action_adds_the_required_predicates_to_the_action_preconditions(depot_nsam: NumericSAMLearner, depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    depot_nsam.update_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if isinstance(precondition, Predicate)]
    preconditions_str = {precondition.untyped_representation for precondition in action_discrete_preconditions}
    assert preconditions_str.issuperset({"(at ?x ?y)"})


def test_handle_single_trajectory_component_does_not_remove_the_required_predicates_to_the_action_preconditions(
    depot_nsam: NumericSAMLearner, depot_observation: Observation
):
    depot_nsam.current_trajectory_objects = depot_observation.grounded_objects
    depot_nsam.handle_single_trajectory_component(depot_observation.components[0])
    learned_action = depot_nsam.partial_domain.actions["drive"]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if isinstance(precondition, Predicate)]
    preconditions_str = {precondition.untyped_representation for precondition in action_discrete_preconditions}
    assert preconditions_str.issuperset({"(at ?x ?y)"})


def test_learn_action_model_returns_learned_model(depot_nsam: NumericSAMLearner, depot_observation: Observation):
    learned_model, learning_metadata = depot_nsam.learn_action_model([depot_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_for_satellite_domain_returns_learned_model(satellite_nsam: NumericSAMLearner, satellite_numeric_observation: Observation):
    learned_model, learning_metadata = satellite_nsam.learn_action_model([satellite_numeric_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_for_satellite_with_problematic_trajectory_domain_returns_learned_model(
    satellite_nsam: NumericSAMLearner, satellite_observation_problematic: Observation
):
    learned_model, learning_metadata = satellite_nsam.learn_action_model([satellite_observation_problematic])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_minecraft_domain_creates_domain_with_correct_preconditions_and_effects(
    minecraft_nsam: NumericSAMLearner, minecraft_observation: Observation
):
    learned_model, learning_metadata = minecraft_nsam.learn_action_model([minecraft_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_minecraft_medium_domain_creates_domain_with_correct_preconditions_and_effects(
    minecraft_medium_sam: NumericSAMLearner, minecraft_medium_observation: Observation
):
    learned_model, learning_metadata = minecraft_medium_sam.learn_action_model([minecraft_medium_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_minecraft_small_domain_creates_domain_with_correct_preconditions_and_effects(
    minecraft_small_sam: NumericSAMLearner, minecraft_small_observation: Observation
):
    learned_model, learning_metadata = minecraft_small_sam.learn_action_model([minecraft_small_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_driverlog_polynomial_domain_returns_non_trivial_polynomial_conditions_to_learned_actions_and_does_not_fail_with_error(
    driverlog_polynomial_nsam: NumericSAMLearner, driverlog_polynomial_observation: Observation
):
    try:
        learned_model, learning_metadata = driverlog_polynomial_nsam.learn_action_model([driverlog_polynomial_observation])
        print()
        print(learning_metadata)
        print(learned_model.to_pddl())
    except Exception as e:
        assert False, f"An error occurred: {e}"


def test_learn_action_model_when_applying_multiple_times_with_different_trajectories_updates_the_learned_action_model_and_allows_learning_when_some_actions_are_deleted_since_they_are_not_being_observed(
    depot_nsam: NumericSAMLearner, depot_observation: Observation
):
    num_triplets = len(depot_observation)
    observation1 = Observation()
    observation2 = Observation()
    observation1.add_problem_objects(depot_observation.grounded_objects)
    observation2.add_problem_objects(depot_observation.grounded_objects)
    for i in range(2):
        observation1.add_component(
            previous_state=depot_observation.components[i].previous_state,
            call=depot_observation.components[i].grounded_action_call,
            next_state=depot_observation.components[i].next_state,
        )

    for i in range(2, num_triplets):
        observation2.add_component(
            previous_state=depot_observation.components[i].previous_state,
            call=depot_observation.components[i].grounded_action_call,
            next_state=depot_observation.components[i].next_state,
        )

    learned_model1, learning_metadata = depot_nsam.learn_action_model([observation1])
    num_learned_actions_model1 = len(learned_model1.actions)
    learned_model2, learning_metadata = depot_nsam.learn_action_model([observation2])
    num_learned_actions_model2 = len(learned_model2.actions)
    assert num_learned_actions_model1 < num_learned_actions_model2


def test_learn_action_model_with_ppo_observations_and_nsam_returns_preconditions_without_duplications_while_using_the_trajectories_incrementally(
    minecraft_ppo_domain: Domain,
):
    minecraft_ppo_nsam = NumericSAMLearner(minecraft_ppo_domain)
    parser = TrajectoryParser(minecraft_ppo_domain)
    learned_model = None
    for observation in TEST_PPO_OBSERVATIONS_DIRECTORY.glob("*.trajectory"):
        trajectory = parser.parse_trajectory(observation)
        learned_model, _ = minecraft_ppo_nsam.learn_action_model([trajectory])

    print()
    print(learned_model.to_pddl())


def test_learn_action_model_with_multiple_farmland_observations_increases_the_size_of_the_dataset_between_iterations(
    farmland_domain: Domain, farmland_nsam: NumericSAMLearner
):
    observations = []
    learned_model = None
    for problem_path in FARMLAND_TRAJECTORIES_DIRECTORY.glob("*.pddl"):
        trajectory_path = FARMLAND_TRAJECTORIES_DIRECTORY / f"{problem_path.stem}.trajectory"
        problem = ProblemParser(problem_path, farmland_domain).parse_problem()
        observation = TrajectoryParser(farmland_domain, problem).parse_trajectory(trajectory_path)
        observations.append(observation)
        learned_model, _ = farmland_nsam.learn_action_model([observation])
        assert len(learned_model.actions) > 0

    one_shot_nsam = NumericSAMLearner(farmland_domain, polynomial_degree=0)
    learned_model_one_shot, _ = one_shot_nsam.learn_action_model(observations)
    assert len(learned_model.actions) == len(learned_model_one_shot.actions)
    # assert learned_model.to_pddl() == learned_model_one_shot.to_pddl()

    print(learned_model_one_shot.to_pddl())
    print(learned_model.to_pddl())


def test_learn_action_model_with_non_numeric_domain_does_not_fail_and_returns_action_model(
    elevators_observation: Observation, elevators_nsam: NumericSAMLearner
):
    try:
        learned_model, _ = elevators_nsam.learn_action_model([elevators_observation])
        print(learned_model.to_pddl())
    except Exception:
        assert False, "An error occurred while learning the action model."



from importlib import resources
from typing import List


def print_domains() -> None:
    """
    List all benchmark domains.
    """
    pkg = f"amlgym.benchmarks.domains"
    print([f.name.split('.')[0] for f in resources.files(pkg).iterdir() if f.is_file()])


def get_domain(domain_name: str) -> str:
    """
    Read the content of a PDDL domain file as text.
    """
    pkg = f"amlgym.benchmarks.domains"
    domain_file = f"{domain_name}.pddl" if '.pddl' not in domain_name else domain_name
    with resources.open_text(pkg, domain_file) as f:
        return f.read()


def get_trajectories(domain_name: str,
                     kind: str = 'learning') -> List[str]:
    """
    Return the absolute path of a PDDL domain trajectory files in the benchmarks.trajectories package.
    """
    possible_kinds = ['learning', 'learning_hard', 'applicability']
    assert kind in possible_kinds, f'`kind` must be one of {possible_kinds}'

    pkg = f"amlgym.benchmarks.trajectories.{kind}.{domain_name.split('.')[0]}"
    trajectories = []
    for traj_file in resources.files(pkg).iterdir():
        with resources.open_text(pkg, traj_file.name) as f:
            trajectories.append(f.read())
    return trajectories


def get_domain_path(domain_name: str) -> str:
    """
    Return the absolute path of a PDDL domain file in the benchmarks.domains package.
    """
    pkg = "amlgym.benchmarks.domains"
    domain_name = f"{domain_name}.pddl" if '.pddl' not in domain_name else domain_name
    # get absolute path of pddl domain file
    domain_path = resources.files(pkg).joinpath(domain_name)
    return str(domain_path)


def get_trajectories_path(domain_name: str,
                          kind: str = 'learning') -> List[str]:
    """
    Return the absolute path of a PDDL domain trajectory files in the benchmarks.trajectories package.
    """
    possible_kinds = ['learning', 'learning_hard', 'applicability']
    assert kind in possible_kinds, f'`kind` must be one of {possible_kinds}'

    pkg = f"amlgym.benchmarks.trajectories.{kind}.{domain_name.split('.')[0]}"
    trajectories_path = [str(f) for f in resources.files(pkg).iterdir() if f.is_file()]
    return sorted(trajectories_path, key=lambda x: int(x.split('/')[-1].split('_')[0]))



"""Module tests the incremental convex hull learner."""

import csv
import random
import time

import numpy
import numpy as np
import pandas as pd
import pytest
from pandas import Series, DataFrame
from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import PDDLFunction, Domain

from sam_learning.core import VocabularyCreator
from sam_learning.core.numeric_learning.numeric_utils import (
    display_convex_hull,
    create_monomials,
    create_polynomial_string,
)
from sam_learning.core.numeric_learning.incremental_convex_hull_learner import IncrementalConvexHullLearner
from tests.consts import (
    FARMLAND_EXAMPLES_PATH,
    FARMLAND_DOMAIN_PATH,
    FARMLAND_PAPER_EXAMPLES_PATH,
    FARMLAND_SPAN_EXAMPLES_PATH,
    DEPOT_ERROR_IN_INCREMENTAL_HULL_CSV,
)

TEST_ACTION_NAME = "test_action"

random.seed(42)


@pytest.fixture
def farmland_domain() -> Domain:
    return DomainParser(FARMLAND_DOMAIN_PATH, partial_parsing=True).parse_domain()


@pytest.fixture
def convex_hull_learner() -> IncrementalConvexHullLearner:
    domain_functions = {
        "x": PDDLFunction(name="x", signature={}),
        "y": PDDLFunction(name="y", signature={}),
        "z": PDDLFunction(name="z", signature={}),
    }
    return IncrementalConvexHullLearner(TEST_ACTION_NAME, domain_functions)


@pytest.fixture
def polynomial_convex_hull_learner() -> IncrementalConvexHullLearner:
    domain_functions = {
        "(x )": PDDLFunction(name="x", signature={}),
        "(y )": PDDLFunction(name="y", signature={}),
        "(z )": PDDLFunction(name="z", signature={}),
    }
    return IncrementalConvexHullLearner(TEST_ACTION_NAME, domain_functions, polynom_degree=1)


def test_creating_the_convex_hull_learner_creates_the_dataframe_with_the_correct_columns_and_zero_rows(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    assert convex_hull_learner.data.shape == (0, 3)
    assert convex_hull_learner.data.columns.tolist() == ["(x )", "(y )", "(z )"]


def test_shift_new_point_when_point_is_single_dimension_series_shifts_point_correctly_and_does_not_fail(
    convex_hull_learner,
):
    point = Series([1])
    convex_hull_learner.affine_independent_data = DataFrame({"(x )": [1]})
    try:
        result = convex_hull_learner._shift_new_point(point)
        assert result is not None
        assert result == numpy.array([[0.0]])
    except Exception as e:
        pytest.fail(f"Exception was raised: {e}")


def test_shift_new_point_when_point_is_single_dimension_dataframe_shifts_point_correctly_and_does_not_fail(
    convex_hull_learner,
):
    point = DataFrame({"(x )": [1]})
    convex_hull_learner.affine_independent_data = DataFrame({"(x )": [1]})
    try:
        result = convex_hull_learner._shift_new_point(point)
        assert result is not None
        assert result == numpy.array([[0.0]])
    except Exception as e:
        pytest.fail(f"Exception was raised: {e}")


def test_add_new_point_when_adding_a_point_for_the_first_time_only_adds_it_to_the_class_data(convex_hull_learner):
    test_sample = {"(x )": 1, "(y )": 2, "(z )": 3}
    convex_hull_learner.add_new_point(test_sample)
    assert convex_hull_learner.data.shape == (1, 3)
    assert convex_hull_learner.data.iloc[0].to_dict() == test_sample
    assert convex_hull_learner._convex_hull is None
    assert convex_hull_learner._gsp_base is None
    assert convex_hull_learner._complementary_base is None
    assert not convex_hull_learner._spanning_standard_base


def test_add_new_point_when_adding_a_point_for_the_second_time_does_not_add_it_again(convex_hull_learner):
    test_sample = {"(x )": 1, "(y )": 2, "(z )": 3}
    convex_hull_learner.add_new_point(test_sample)
    convex_hull_learner.add_new_point(test_sample)
    assert convex_hull_learner.data.shape == (1, 3)


def test_add_new_point_when_adding_two_points_creates_a_base_for_the_points_and_an_orthonormal_complementary_base(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 1, "(y )": 2, "(z )": 3}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 2, "(y )": 3, "(z )": 4}
    convex_hull_learner.add_new_point(second_sample)
    assert convex_hull_learner.data.shape == (2, 3)
    assert convex_hull_learner.data.iloc[0].to_dict() == first_sample
    assert convex_hull_learner.data.iloc[1].to_dict() == second_sample
    assert convex_hull_learner._convex_hull is None
    assert convex_hull_learner._gsp_base is not None
    assert len(convex_hull_learner.additional_dependency_conditions) == 2


def test_add_new_point_when_adding_three_points_with_that_create_larger_base_changes_the_base_created_by_the_algorithm(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    assert convex_hull_learner.data.shape == (2, 3)
    assert convex_hull_learner.data.iloc[0].to_dict() == first_sample
    assert convex_hull_learner.data.iloc[1].to_dict() == second_sample
    normal_base = convex_hull_learner._gsp_base.copy()
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    assert convex_hull_learner.data.shape == (3, 3)
    assert len(convex_hull_learner._gsp_base) == len(normal_base) + 1


def test_add_new_point_when_adding_three_points_when_gsp_returns_two_dimensions_sets_the_convex_hull_with_enough_data(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    assert convex_hull_learner._convex_hull is not None
    assert len(convex_hull_learner._convex_hull.points) == 3
    display_convex_hull(TEST_ACTION_NAME, True, convex_hull_learner._convex_hull)


def test_add_new_point_when_adding_four_points_and_the_last_one_is_linearly_dependent_does_not_change_the_base(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    normal_base = convex_hull_learner._gsp_base.copy()
    fourth_sample = {"(x )": 2, "(y )": 2, "(z )": 0}
    convex_hull_learner.add_new_point(fourth_sample)
    assert normal_base == convex_hull_learner._gsp_base


def test_add_new_point_when_adding_four_points_and_the_last_one_is_linearly_dependent_adds_the_point_to_the_convex_hull(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    fourth_sample = {"(x )": 2, "(y )": 2, "(z )": 0}
    convex_hull_learner.add_new_point(fourth_sample)
    assert len(convex_hull_learner._convex_hull.points) == 4
    display_convex_hull(TEST_ACTION_NAME, True, convex_hull_learner._convex_hull)


def test_add_new_point_when_adding_one_point_at_a_time_returns_correct_values_for_convex_hull_points_without_subtracting_first_sample(
    farmland_domain: Domain,
):
    test_action_name = "move-slow"
    dataframe = pd.read_csv(FARMLAND_SPAN_EXAMPLES_PATH)
    domain_functions = VocabularyCreator().create_lifted_functions_vocabulary(
        domain=farmland_domain, possible_parameters=farmland_domain.actions[test_action_name].signature
    )

    learner = IncrementalConvexHullLearner(
        test_action_name, domain_functions=domain_functions, relevant_fluents=["(x ?f1)", "(x ?f2)"]
    )
    for index, (_, row) in enumerate(dataframe.iterrows()):
        if learner._spanning_standard_base:
            assert (learner._convex_hull.points == (dataframe[:index].to_numpy(dtype=numpy.float32))).all()

        point_data = {key: row[key] for key in dataframe.columns.tolist()}
        learner.add_new_point(point_data)


def test_add_new_point_when_adding_a_point_with_feature_not_existing_in_previous_sample_does_not_add_new_feature_to_dataset(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 1, "(w )": 1}
    convex_hull_learner.add_new_point(second_sample)
    assert convex_hull_learner.data.shape[0] == 2
    assert convex_hull_learner.data.shape[1] == 3


def test_add_new_point_when_adding_multiple_points_with_single_dimension_does_not_try_to_create_a_convex_hull(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    convex_hull_learner.relevant_fluents = ["(x )"]
    try:
        for i in range(10):
            sample = {"(x )": i}
            convex_hull_learner.add_new_point(sample)

    except Exception as e:
        pytest.fail(f"Exception was raised: {e}")


def test_add_new_point_when_adding_four_points_that_span_the_entire_space_returns_standard_basis(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    fourth_sample = {"(x )": 0, "(y )": 0, "(z )": 1}
    convex_hull_learner.add_new_point(fourth_sample)
    assert len(convex_hull_learner._convex_hull.points) == 4
    assert convex_hull_learner._gsp_base == [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    display_convex_hull(TEST_ACTION_NAME, True, convex_hull_learner._convex_hull)


def test_add_new_point_when_relevant_features_is_empty_does_not_compute_conditions_basis_or_convex_hull(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    convex_hull_learner.relevant_fluents = []
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 4}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 12}
    convex_hull_learner.add_new_point(third_sample)
    fourth_sample = {"(x )": 0, "(y )": 0, "(z )": 1}
    convex_hull_learner.add_new_point(fourth_sample)
    assert convex_hull_learner._convex_hull is None
    assert convex_hull_learner._gsp_base is None
    assert convex_hull_learner._complementary_base is None
    assert convex_hull_learner.additional_dependency_conditions == []
    assert convex_hull_learner.affine_independent_data.empty


def test_add_new_point_when_relevant_features_with_less_features_than_original_returns_convex_hull_on_correct_dimensions(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    convex_hull_learner.relevant_fluents = ["(y )", "(x )"]
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 4}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 12}
    convex_hull_learner.add_new_point(third_sample)
    fourth_sample = {"(x )": 0, "(y )": 0, "(z )": 1}
    convex_hull_learner.add_new_point(fourth_sample)
    assert len(convex_hull_learner._convex_hull.points) == 4
    assert convex_hull_learner._spanning_standard_base is True
    display_convex_hull(TEST_ACTION_NAME, True, convex_hull_learner._convex_hull)


def test_add_new_point_when_relevant_features_with_less_features_than_original_and_feature_not_in_relevant_fluents_is_constant_not_creating_conditions_for_it(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    convex_hull_learner.relevant_fluents = ["(y )", "(x )"]
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    fourth_sample = {"(x )": 0, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(fourth_sample)
    assert len(convex_hull_learner.additional_dependency_conditions) == 0


def test_incremental_create_ch_inequalities_with_one_dimension_returns_min_max_conditions(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    (coefficients, border_point, transformed_vars, _) = convex_hull_learner._incremental_create_ch_inequalities(
        display_mode=True
    )
    assert coefficients == [[-1], [1]]
    assert border_point == [0.0, 1.0]
    assert transformed_vars == ["(* (- (y ) 1) -1)"]
    assert set(convex_hull_learner.additional_dependency_conditions) == {"(= (z ) 0)", "(= (x ) (+ (* (y ) -1) 1))"}


def test_incremental_create_ch_inequalities_with_one_dimension_returns_min_max_conditions_and_correct_complementary_conditions(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 3, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    fourth_sample = {"(x )": 4, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(fourth_sample)
    (
        coefficients,
        border_point,
        transformed_vars,
        span_verification_conditions,
    ) = convex_hull_learner._incremental_create_ch_inequalities(display_mode=True)
    assert coefficients == [[-1], [1]]
    assert border_point == [0, 4]
    assert set(convex_hull_learner.additional_dependency_conditions) == {"(= (y ) 1)", "(= (z ) 0)"}
    assert len(span_verification_conditions) == 0
    assert transformed_vars == ["(x )"]


def test_incremental_create_ch_inequalities_with_point_spanning_standard_base_returns_correct_convex_hull_with_no_verification_conditions(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 0, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(first_sample)
    second_sample = {"(x )": 1, "(y )": 0, "(z )": 0}
    convex_hull_learner.add_new_point(second_sample)
    third_sample = {"(x )": 1, "(y )": 1, "(z )": 0}
    convex_hull_learner.add_new_point(third_sample)
    fourth_sample = {"(x )": 0, "(y )": 0, "(z )": 1}
    convex_hull_learner.add_new_point(fourth_sample)
    (
        coefficients,
        border_point,
        transformed_vars,
        span_verification_conditions,
    ) = convex_hull_learner._incremental_create_ch_inequalities(display_mode=True)
    assert len(coefficients) == 4
    assert len(border_point) == 4
    assert span_verification_conditions is None


def test_construct_safe_linear_inequalities_when_the_number_of_samples_is_one_creates_a_single_condition(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    first_sample = {"(x )": 2, "(y )": 3, "(z )": -1}
    convex_hull_learner.add_new_point(first_sample)
    inequality_precondition = convex_hull_learner.construct_safe_linear_inequalities()
    assert inequality_precondition.binary_operator == "and"
    assert len(inequality_precondition.operands) == 3
    assert {op.to_pddl() for op in inequality_precondition.operands} == {
        "(= (x ) 2)",
        "(= (y ) 3)",
        "(= (z ) -1)",
    }


def test_construct_safe_linear_inequalities_when_creating_convex_hull_with_large_number_of_samples_does_not_fail(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    for _ in range(100):
        convex_hull_learner.add_new_point(
            {"(x )": random.uniform(-100, 100), "(y )": random.uniform(-100, 100), "(z )": random.uniform(-100, 100)}
        )

    inequality_precondition = convex_hull_learner.construct_safe_linear_inequalities()
    assert inequality_precondition.binary_operator == "and"
    assert len(convex_hull_learner._convex_hull.points) == 100
    assert convex_hull_learner._gsp_base == [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    assert convex_hull_learner._complementary_base == []
    display_convex_hull(TEST_ACTION_NAME, True, convex_hull_learner._convex_hull)


def test_construct_safe_linear_inequalities_when_adding_extra_large_number_of_samples_does_not_take_extremely_long_to_calculate(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    start_time = time.time()
    for _ in range(1000):
        convex_hull_learner.add_new_point(
            {"(x )": random.uniform(-100, 100), "(y )": random.uniform(-100, 100), "(z )": random.uniform(-100, 100)}
        )

    inequality_precondition = convex_hull_learner.construct_safe_linear_inequalities()
    end_time = time.time()
    assert end_time - start_time < 5 * 60
    assert inequality_precondition.binary_operator == "and"
    assert len(convex_hull_learner._convex_hull.points) == 1000
    assert convex_hull_learner._gsp_base == [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    assert convex_hull_learner._complementary_base == []
    display_convex_hull(TEST_ACTION_NAME, True, convex_hull_learner._convex_hull)


def test_incremental_create_ch_inequalities_creates_convex_hull_with_all_variables_in_the_vectors_including_polynoms(
    polynomial_convex_hull_learner: IncrementalConvexHullLearner,
):
    k = 50
    numbers = list(range(100))
    state_matrix = {
        "(x )": random.sample(numbers, k=k),
        "(y )": random.sample(numbers, k=k),
        "(z )": random.sample(numbers, k=k),
    }
    monomials = create_monomials(list(state_matrix.keys()), polynom_degree=1)
    for i in range(k):
        polynomial_convex_hull_learner.add_new_point(
            {
                create_polynomial_string(monomial): np.prod([state_matrix[name][i] for name in monomial])
                for monomial in monomials
            },
        )

    A, _, _, _ = polynomial_convex_hull_learner._incremental_create_ch_inequalities()
    assert all([len(item) == 6 for item in A])


def test_construct_convex_hull_inequalities_when_given_polynomial_inequalities_returns_convex_hull_with_all_points_as_conditions(
    polynomial_convex_hull_learner: IncrementalConvexHullLearner,
):
    k = 50
    numbers = list(range(100))
    state_matrix = {
        "(x )": random.sample(numbers, k=k),
        "(y )": random.sample(numbers, k=k),
        "(z )": random.sample(numbers, k=k),
    }
    monomials = create_monomials(list(state_matrix.keys()), polynom_degree=1)
    for i in range(k):
        polynomial_convex_hull_learner.add_new_point(
            {
                create_polynomial_string(monomial): np.prod([state_matrix[name][i] for name in monomial])
                for monomial in monomials
            },
        )

    precondition = polynomial_convex_hull_learner.construct_safe_linear_inequalities()
    print(str(precondition))


def test_construct_convex_hull_inequalities_when_adding_multiple_points_with_single_dimension_creates_min_max_conditions_and_does_not_raise_an_error(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    convex_hull_learner.relevant_fluents = ["(x )"]
    try:
        for i in range(10):
            sample = {"(x )": i}
            convex_hull_learner.add_new_point(sample)

        precondition = convex_hull_learner.construct_safe_linear_inequalities()
        assert precondition.binary_operator == "and"
        assert len(precondition.operands) == 2
        assert {op.to_pddl() for op in precondition.operands} == {"(>= (x ) 0)", "(<= (x ) 9)"}

    except Exception as e:
        pytest.fail(f"Exception was raised: {e}")


def test_construct_convex_hull_inequalities_when_spanning_standard_basis_returns_correct_number_of_equations(
    farmland_domain: Domain,
):
    dataframe = pd.read_csv(FARMLAND_SPAN_EXAMPLES_PATH)
    vocabulary_creator = VocabularyCreator()
    possible_bounded_functions = vocabulary_creator.create_lifted_functions_vocabulary(
        domain=farmland_domain, possible_parameters=farmland_domain.actions["move-slow"].signature
    )
    learner = IncrementalConvexHullLearner(
        "move-slow", domain_functions=possible_bounded_functions, relevant_fluents=["(x ?f1)", "(x ?f2)"]
    )
    for index, (_, row) in enumerate(dataframe.iterrows()):

        if learner._spanning_standard_base:
            assert (learner._convex_hull.points == (dataframe[:index].to_numpy(dtype=numpy.float32))).all()

        point_data = {key: row[key] for key in dataframe.columns.tolist()}
        learner.add_new_point(point_data)

    precondition = learner.construct_safe_linear_inequalities()
    assert len(precondition.operands) == 4


def test_construct_convex_hull_inequalities_when_spanning_standard_basis_returns_with_one_constant_variable(
    farmland_domain: Domain,
):
    dataframe = pd.read_csv(FARMLAND_EXAMPLES_PATH)
    vocabulary_creator = VocabularyCreator()
    possible_bounded_functions = vocabulary_creator.create_lifted_functions_vocabulary(
        domain=farmland_domain, possible_parameters=farmland_domain.actions["move-slow"].signature
    )
    learner = IncrementalConvexHullLearner("move-slow", domain_functions=possible_bounded_functions)
    for index, (_, row) in enumerate(dataframe.iterrows()):
        if learner._spanning_standard_base:
            break

        point_data = {key: row[key] for key in dataframe.columns.tolist()}
        learner.add_new_point(point_data)

    assert len(learner._gsp_base) == 2
    assert set(learner.additional_dependency_conditions) == {"(= (x ?f2) 0)"}
    assert len(learner._convex_hull.points) == 6


def test_construct_convex_hull_inequalities_when_not_spanning_standard_basis_returns_with_two_constant_variables(
    farmland_domain: Domain,
):
    dataframe = pd.read_csv(FARMLAND_EXAMPLES_PATH)[:4]
    vocabulary_creator = VocabularyCreator()
    possible_bounded_functions = vocabulary_creator.create_lifted_functions_vocabulary(
        domain=farmland_domain, possible_parameters=farmland_domain.actions["move-slow"].signature
    )
    learner = IncrementalConvexHullLearner("move-slow", domain_functions=possible_bounded_functions)
    for index, (_, row) in enumerate(dataframe.iterrows()):
        if learner._spanning_standard_base:
            assert len(learner._gsp_base) == 1
            assert learner._convex_hull.points is None

        point_data = {key: row[key] for key in dataframe.columns.tolist()}
        learner.add_new_point(point_data)

    assert len(learner._gsp_base) == 1
    assert set(learner.additional_dependency_conditions) == {"(= (cost ) 1)", "(= (x ?f2) 0)"}


def test_construct_convex_hull_inequalities_when_given_too_few_examples_returns_lower_dimensional_convex_hull_with_real_dataset_example(
    farmland_domain: Domain,
):
    dataframe = pd.read_csv(FARMLAND_PAPER_EXAMPLES_PATH)
    vocabulary_creator = VocabularyCreator()
    possible_bounded_functions = vocabulary_creator.create_lifted_functions_vocabulary(
        domain=farmland_domain, possible_parameters=farmland_domain.actions["move-slow"].signature
    )
    learner = IncrementalConvexHullLearner("move-slow", domain_functions=possible_bounded_functions)
    for index, (_, row) in enumerate(dataframe.iterrows()):
        point_data = {key: row[key] for key in dataframe.columns.tolist()}
        learner.add_new_point(point_data)

    assert learner._gsp_base is not None
    assert learner._complementary_base is not None
    precondition = learner.construct_safe_linear_inequalities()
    print(precondition.print(decimal_digits=2))


def test_construct_convex_hull_inequalities_when_convex_hull_initialized_with_empty_relevant_fluents_returns_empty_conditions(
    farmland_domain: Domain,
):
    test_action_name = "move-slow"
    dataframe = pd.read_csv(FARMLAND_SPAN_EXAMPLES_PATH)
    domain_functions = VocabularyCreator().create_lifted_functions_vocabulary(
        domain=farmland_domain, possible_parameters=farmland_domain.actions[test_action_name].signature
    )

    learner = IncrementalConvexHullLearner(test_action_name, domain_functions=domain_functions, relevant_fluents=[])
    for index, (_, row) in enumerate(dataframe.iterrows()):
        point_data = {key: row[key] for key in dataframe.columns.tolist()}
        point_data["(cost )"] = 0
        learner.add_new_point(point_data)

    precondition = learner.construct_safe_linear_inequalities()
    assert len(precondition.operands) == 0


def test_reset_when_creating_one_feature_relevant_creates_conditions_with_only_one_variables(
    polynomial_convex_hull_learner: IncrementalConvexHullLearner,
):
    k = 50
    numbers = list(range(100))
    state_matrix = {
        "(x )": random.sample(numbers, k=k),
        "(y )": random.sample(numbers, k=k),
        "(z )": random.sample(numbers, k=k),
    }
    monomials = create_monomials(list(state_matrix.keys()), polynom_degree=1)
    for i in range(k):
        polynomial_convex_hull_learner.add_new_point(
            {
                create_polynomial_string(monomial): np.prod([state_matrix[name][i] for name in monomial])
                for monomial in monomials
            },
        )

    precondition = polynomial_convex_hull_learner.construct_safe_linear_inequalities()
    print(str(precondition))
    assert "(x )" in str(precondition)
    assert "(y )" in str(precondition)
    assert "(z )" in str(precondition)
    polynomial_convex_hull_learner.reset(["(x )"])
    precondition = polynomial_convex_hull_learner.construct_safe_linear_inequalities()
    print()
    print(str(precondition))
    assert "(x )" in str(precondition)
    assert "(y )" not in str(precondition)
    assert "(z )" not in str(precondition)


def test_reset_followed_by_add_point_when_point_outside_range_adds_point_correctly_and_does_not_fail(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    for i in range(10):
        convex_hull_learner.add_new_point(
            {"(x )": random.uniform(0, 100), "(y )": random.uniform(0, 100), "(z )": random.uniform(0, 100)}
        )

    convex_hull_learner.reset(["(x )"])
    convex_hull_learner.add_new_point({"(x )": 200, "(y )": 500, "(z )": 120})
    precondition = convex_hull_learner.construct_safe_linear_inequalities()
    print(str(precondition))
    assert len(precondition.operands) == 2
    assert "(<= (x ) 200)" in str(precondition)


def test_is_point_in_convex_hull_captures_that_a_point_is_in_a_convex_hull_in_a_2d_plane(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    valid_x_points = [0, 0, 1, 1]
    valid_y_points = [0, 1, 0, 1]
    for x, y in zip(valid_x_points, valid_y_points):
        convex_hull_learner.add_new_point({"(x )": x, "(y )": y, "(z )": 0})

    point_to_test = pd.Series({"(x )": 0.5, "(y )": 0.5, "(z )": 0})
    assert convex_hull_learner.is_point_in_convex_hull(point_to_test)


def test_is_point_in_convex_hull_captures_that_a_point_is_not_in_a_convex_hull_in_a_2d_plane(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    valid_x_points = [0, 0, 1, 1]
    valid_y_points = [0, 1, 0, 1]
    for x, y in zip(valid_x_points, valid_y_points):
        convex_hull_learner.add_new_point({"(x )": x, "(y )": y, "(z )": 0})

    point_to_test = pd.Series({"(x )": 2, "(y )": 2, "(z )": 0})
    assert not convex_hull_learner.is_point_in_convex_hull(point_to_test)


def test_is_point_in_convex_hull_captures_that_all_given_points_are_in_the_convex_hull_when_running_with_multiple_inputs_conscutively(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    valid_x_points = [0, 0, 1, 1]
    valid_y_points = [0, 1, 0, 1]
    for x, y in zip(valid_x_points, valid_y_points):
        convex_hull_learner.add_new_point({"(x )": x, "(y )": y, "(z )": 0})

    for i in range(10):
        point_to_test = pd.Series({"(x )": random.uniform(0, 1), "(y )": random.uniform(0, 1), "(z )": 0})
        assert convex_hull_learner.is_point_in_convex_hull(point_to_test)


def test_is_point_in_convex_hull_captures_that_more_than_one_point_is_in_1d_convex_hull(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    for i in range(10):
        convex_hull_learner.add_new_point({"(x )": i, "(y )": 0, "(z )": 0})

    point_to_test = pd.Series({"(x )": random.uniform(0, 10), "(y )": 0, "(z )": 0})
    assert convex_hull_learner.is_point_in_convex_hull(point_to_test)


def test_copy_returns_correct_convex_hull_when_data_is_empty(convex_hull_learner: IncrementalConvexHullLearner):
    convex_hull_learner_copy = convex_hull_learner.copy()
    assert convex_hull_learner_copy._convex_hull is None
    assert convex_hull_learner_copy._gsp_base is None
    assert convex_hull_learner_copy._complementary_base is None
    assert not convex_hull_learner_copy._spanning_standard_base
    assert convex_hull_learner_copy._convex_hull is None


def test_copy_returns_correct_convex_hull_when_data_is_not_empty(convex_hull_learner: IncrementalConvexHullLearner):
    valid_x_points = [0, 0, 1, 1]
    valid_y_points = [0, 1, 0, 1]
    for x, y in zip(valid_x_points, valid_y_points):
        convex_hull_learner.add_new_point({"(x )": x, "(y )": y, "(z )": 0})

    copy = convex_hull_learner.copy()
    assert copy._convex_hull is not None
    assert copy._gsp_base == convex_hull_learner._gsp_base
    assert copy._complementary_base == convex_hull_learner._complementary_base
    assert copy._spanning_standard_base == convex_hull_learner._spanning_standard_base
    assert np.array_equal(copy._convex_hull.points, convex_hull_learner._convex_hull.points)


def test_copy_returns_correct_convex_hull_when_data_is_not_empty_and_large_number_of_points_were_added_to_convex_hull(
    convex_hull_learner: IncrementalConvexHullLearner,
):
    for i in range(100):
        convex_hull_learner.add_new_point(
            {"(x )": random.uniform(0, 100), "(y )": random.uniform(0, 100), "(z )": random.uniform(0, 100)}
        )

    copy = convex_hull_learner.copy()
    assert copy._convex_hull is not None
    assert copy._gsp_base == convex_hull_learner._gsp_base
    assert copy._complementary_base == convex_hull_learner._complementary_base
    assert copy._spanning_standard_base == convex_hull_learner._spanning_standard_base
    assert np.array_equal(copy._convex_hull.points, convex_hull_learner._convex_hull.points)


def test_depot_production_error_data_should_not_cause_exception(depot_domain: Domain):
    # This test is to ensure that the data from the depot production error does not cause an exception
    # when added to the convex hull learner.
    function_vocabulary_creator = VocabularyCreator()
    test_action = "unload"
    pb_functions = function_vocabulary_creator.create_lifted_functions_vocabulary(
        domain=depot_domain, possible_parameters=depot_domain.actions[test_action].signature
    )
    learner = IncrementalConvexHullLearner(test_action, pb_functions, qhull_options="QJ Q11 Q12")

    with open(DEPOT_ERROR_IN_INCREMENTAL_HULL_CSV, "rt") as csv_file:
        csv_reader = csv.DictReader(csv_file)
        for row in csv_reader:
            test_data = {key: float(value) for key, value in row.items() if key != "time"}
            try:
                learner.add_new_point(test_data)
            except Exception as e:
                pytest.fail(f"Exception was raised while adding data: {e}")


def test_depot_production_error_data_should_not_cause_exception_without_qhull_options(depot_domain: Domain):
    # This test is to ensure that the data from the depot production error does not cause an exception
    # when added to the convex hull learner.
    function_vocabulary_creator = VocabularyCreator()
    test_action = "unload"
    pb_functions = function_vocabulary_creator.create_lifted_functions_vocabulary(
        domain=depot_domain, possible_parameters=depot_domain.actions[test_action].signature
    )
    learner = IncrementalConvexHullLearner(test_action, pb_functions)

    with open(DEPOT_ERROR_IN_INCREMENTAL_HULL_CSV, "rt") as csv_file:
        csv_reader = csv.DictReader(csv_file)
        for row in csv_reader:
            test_data = {key: float(value) for key, value in row.items() if key != "time"}
            try:
                learner.add_new_point(test_data)
            except Exception as e:
                pytest.fail(f"Exception was raised while adding data: {e}")



"""Extension to SAM Learning that can learn numeric state variables."""

from typing import List, Dict, Tuple, Optional

from pddl_plus_parser.models import Observation, ActionCall, State, Domain, Precondition, NumericalExpressionTree

from sam_learning.core import (
    LearnerDomain,
    NumericFluentStateStorage,
    NumericFunctionMatcher,
    NotSafeActionError,
    LearnerAction,
)
from sam_learning.core.learner_domain import DISJUNCTIVE_PRECONDITIONS_REQ
from sam_learning.learners.sam_learning import SAMLearner
from utilities import NegativePreconditionPolicy


class NumericSAMLearner(SAMLearner):
    """The Extension of SAM that is able to learn numeric state variables."""

    storage: Dict[str, NumericFluentStateStorage]
    function_matcher: NumericFunctionMatcher
    relevant_fluents: Dict[str, List[str]]

    def __init__(
        self,
        partial_domain: Domain,
        relevant_fluents: Optional[Dict[str, List[str]]] = None,
        allow_unsafe: bool = False,
        polynomial_degree: int = 0,
        negative_preconditions_policy: NegativePreconditionPolicy = NegativePreconditionPolicy.no_remove,
        epsilon: float = 0.0,
        qhull_options: str = "",
        **kwargs,
    ):
        super().__init__(partial_domain, negative_preconditions_policy=negative_preconditions_policy)
        self.storage = {}
        self.function_matcher = NumericFunctionMatcher(partial_domain)
        self.relevant_fluents = relevant_fluents
        self._allow_unsafe = allow_unsafe
        self.polynom_degree = polynomial_degree
        self._epsilon = epsilon
        self._qhull_options = qhull_options

    def _remove_deprecated_numeric_preconditions(self, action: LearnerAction) -> None:
        """Removes the deprecated numeric preconditions from the input action (used when the learning process is iterative).

        :param action: the action that the preconditions are removed from.
        """
        self.logger.debug(f"Removing the deprecated numeric preconditions from the action - {action.name}.")
        non_numeric_preconditions = Precondition("and")
        for condition in action.preconditions.root.operands:
            if not isinstance(condition, NumericalExpressionTree):
                non_numeric_preconditions.add_condition(condition)

        action.preconditions.root = non_numeric_preconditions

    def _construct_safe_numeric_preconditions(self, action: LearnerAction) -> None:
        """Constructs the safe preconditions for the input action.

        :param action: the action that the preconditions are constructed for.
        """
        action_name = action.name
        self._remove_deprecated_numeric_preconditions(action)
        if self.relevant_fluents is None:
            learned_numeric_preconditions = self.storage[action_name].construct_safe_linear_inequalities()

        elif len(self.relevant_fluents[action_name]) == 0:
            self.logger.debug(f"The action {action_name} has no numeric preconditions.")
            return

        else:
            learned_numeric_preconditions = self.storage[action_name].construct_safe_linear_inequalities(self.relevant_fluents[action_name])

        if learned_numeric_preconditions.binary_operator == "and":
            self.logger.debug("The learned preconditions are a conjunction. Adding the internal numeric conditions.")
            for condition in learned_numeric_preconditions.operands:
                action.preconditions.add_condition(condition)

            return

        self.logger.debug("The learned preconditions are not a conjunction. Adding them as a separate condition.")
        action.preconditions.add_condition(learned_numeric_preconditions)
        self.partial_domain.requirements.append(DISJUNCTIVE_PRECONDITIONS_REQ)

    def _construct_safe_numeric_effects(self, action: LearnerAction) -> bool:
        """Constructs the safe numeric effects for the input action.

        :param action: the action that its effects are constructed for.
        :return: whether the effects were learned perfectly.
        """
        if self.relevant_fluents is not None and len(self.relevant_fluents[action.name]) == 0:
            self.logger.debug(f"The action - {action.name} has no numeric effects.")
            action.numeric_effects = set()
            return True

        effects, numeric_preconditions, learned_perfectly = self.storage[action.name].construct_assignment_equations(
            allow_unsafe=self._allow_unsafe, relevant_fluents=self.relevant_fluents[action.name] if self.relevant_fluents is not None else None
        )
        if effects is not None and len(effects) > 0:
            action.numeric_effects = effects

        else:
            self.logger.debug(f"The action - {action.name} has no numeric effects.")
            action.numeric_effects = set()

        self.logger.debug(f"No feature selection applied, using the numeric preconditions as is.")
        return learned_perfectly

    def add_new_action(self, grounded_action: ActionCall, previous_state: State, next_state: State, incremental: bool = False) -> None:
        """Adds a new action to the learned domain.

        :param grounded_action: the grounded action that was executed according to the observation.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous state.
        :param incremental: whether the learning process is incremental or not.
        """
        super().add_new_action(grounded_action, previous_state, next_state)
        self.logger.debug(f"Creating the new storage for the action - {grounded_action.name}.")
        previous_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.previous_state_functions)
        next_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.next_state_functions)
        possible_bounded_functions = self.vocabulary_creator.create_lifted_functions_vocabulary(
            domain=self.partial_domain, possible_parameters=self.partial_domain.actions[grounded_action.name].signature
        )
        self.storage[grounded_action.name] = NumericFluentStateStorage(
            action_name=grounded_action.name,
            domain_functions=possible_bounded_functions,
            polynom_degree=self.polynom_degree,
            epsilon=self._epsilon,
            qhull_options=self._qhull_options,
            incremental=incremental,
        )
        self.storage[grounded_action.name].add_to_previous_state_storage(previous_state_lifted_matches)
        self.storage[grounded_action.name].add_to_next_state_storage(next_state_lifted_matches)
        self.logger.debug(f"Done creating the numeric state variable storage for the action - {grounded_action.name}")

    def update_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Updates the action's data according to the new input observed triplet.

        :param grounded_action: the grounded action that was observed.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        action_name = grounded_action.name
        super().update_action(grounded_action, previous_state, next_state)
        self.logger.debug(f"Adding the numeric state variables to the numeric storage of action - {action_name}.")
        previous_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.previous_state_functions)
        next_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.next_state_functions)
        self.storage[action_name].add_to_previous_state_storage(previous_state_lifted_matches)
        self.storage[action_name].add_to_next_state_storage(next_state_lifted_matches)
        self.logger.debug(f"Done updating the numeric state variable storage for the action - {grounded_action.name}")

    def _create_safe_action(self, action_name: str) -> LearnerAction:
        """Creates a safe action that can be executed in the environment.

        :param action_name: the name of the action to create.
        :return: the safe action that can be executed in the environment.
        """
        action = self.partial_domain.actions[action_name]
        self._construct_safe_numeric_preconditions(action)
        self._construct_safe_numeric_effects(action)
        self.logger.info(f"Done learning the action - {action_name}!")
        return action

    def _create_safe_action_model(self) -> Tuple[Dict[str, LearnerAction], Dict[str, str]]:
        """Crates the safe action model learned by the algorithm.

        :return: the actions that are allowed to execute and the metadata about the learning.
        """
        allowed_actions = {}
        learning_metadata = {}
        for action_name, action in self.partial_domain.actions.items():
            if action_name not in self.storage:
                self.logger.debug(f"The action - {action_name} has not been observed in the trajectories!")
                continue

            try:
                safe_action = self._create_safe_action(action_name)
                allowed_actions[action_name] = safe_action
                learning_metadata[action_name] = "OK"

            except NotSafeActionError as e:
                self.logger.debug(f"The action - {e.action_name} is not safe for execution, reason - {e.reason}")
                action.preconditions = Precondition("and")
                action.discrete_effects = set()
                action.numeric_effects = set()
                learning_metadata[action_name] = e.solution_type.name

        self.partial_domain.actions = allowed_actions
        return allowed_actions, learning_metadata

    def learn_action_model(self, observations: List[Observation]) -> Tuple[LearnerDomain, Dict[str, str]]:
        """Learn the SAFE action model from the input observations.

        :param observations: the list of trajectories that are used to learn the safe action model.
        :return: a domain containing the actions that were learned and the metadata about the learning.
        """
        self.logger.info("Starting to learn the action model!")
        if len(self.partial_domain.functions) == 0:
            self.logger.info(f"The domain {self.partial_domain.name} does not have any numeric state variables.")
            return super().learn_action_model(observations)

        # The domain has numeric state variables.
        super().start_measure_learning_time()
        super().deduce_initial_inequality_preconditions()
        self._complete_possibly_missing_actions()
        for observation in observations:
            self.current_trajectory_objects = observation.grounded_objects
            for component in observation.components:
                if not component.is_successful:
                    self.logger.warning("Skipping the transition because it was not successful.")
                    continue

                self.handle_single_trajectory_component(component)

        self.handle_negative_preconditions_policy()
        allowed_actions, learning_metadata = self._create_safe_action_model()

        super().end_measure_learning_time()
        learning_metadata["learning_time"] = str(self.learning_end_time - self.learning_start_time)
        return self.partial_domain, learning_metadata



"""Auxiliary functions to help the matching process."""
from itertools import permutations
from typing import List, Tuple, Set

from pddl_plus_parser.models import GroundedPredicate


def create_signature_permutations(call_parameters: List[str], lifted_signature: List[str], subset_size: int) -> List[Tuple[Tuple[str]]]:
    """Choose r items our of a list size n.

    :param call_parameters: the parameters in which the action was called with.
    :param lifted_signature: the action's lifted signature definition including possible constants.
    :param subset_size: the size of the subset.
    :return: a list containing matches of grounded objects and their lifted action parameter name counterparts.
    """
    matching_signatures = zip(call_parameters, list(lifted_signature))
    matching_permutations = list(permutations(matching_signatures, subset_size))
    return matching_permutations


def contains_duplicates(parameter_objects: List[str]) -> bool:
    """Validate that the predicate has only one possible match in the literal.

    :param parameter_objects: the of objects to test if contains duplicates.
    :return: whether there are duplicates.
    """
    return len(set(parameter_objects)) != len(parameter_objects)


def extract_discrete_effects(
    previous_state: Set[GroundedPredicate], next_state: Set[GroundedPredicate]
) -> Tuple[Set[GroundedPredicate], Set[GroundedPredicate]]:
    """Extracts discrete the effects from the state object.

    :param previous_state: the previous state object containing the grounded literals.
    :param next_state: the next state object containing its grounded literals.
    :return: tuple containing the add and delete effects of the action.
    """
    add_effects = set([predicate.copy() for predicate in next_state.difference(previous_state) if predicate.is_positive])
    delete_effects = set([predicate.copy(is_negated=True) for predicate in previous_state.difference(next_state) if predicate.is_positive])
    return add_effects, delete_effects


def extract_not_effects(next_state_predicate: Set[GroundedPredicate]) -> Set[GroundedPredicate]:
    """Extracts all negations of grounded predicates that were in post state

    :param next_state_predicate: all the grounded literals in the next state (negative and positive predicates).
    :return: A tuple representing the predicates that cannot be add-effects and those that cannot be delete-effects.
    """
    return {predicate.copy(is_negated=predicate.is_positive) for predicate in next_state_predicate}



"""Creates the trajectories that will be used in the trajectory"""
import logging
import random
import shutil
import sys
from pathlib import Path
from typing import Dict, Set, List, Tuple, Optional

from pddl_plus_parser.exporters import TrajectoryExporter, TrajectoryTriplet
from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser
from pddl_plus_parser.models import PDDLObject, Domain, ActionCall, Problem, State, Operator

from sam_learning.core import VocabularyCreator
from utilities import SolverType

MAX_NUM_STEPS_IN_TRAJECTORY = 200
inapplicable_action_probability = 0.25
random.seed(42)


class RandomWalkTrajectoriesCreator:
    """Class that creates trajectories from executing random walks in the domain."""

    domain_file_name: str
    working_directory_path: Path
    vocabulary_creator: VocabularyCreator
    logger: logging.Logger

    def __init__(self, domain_file_name: str, working_directory_path: Path):
        self.domain_file_name = domain_file_name
        self.working_directory_path = working_directory_path
        self.vocabulary_creator = VocabularyCreator()
        self.logger = logging.getLogger(__name__)

    def _select_inapplicable_action(
        self, domain: Domain, problem: Problem, current_state: State, ground_actions: Set[ActionCall]
    ) -> Tuple[Operator, State]:
        """Selects an inapplicable action for the current state (This is for performance evaluation purposes).

        :param domain: the domain to select the inapplicable action from.
        :param problem: the problem defining the objects and the initial state.
        :param current_state: the current state to select the inapplicable action for.
        :param ground_actions: the grounded actions to select the inapplicable action from.
        :return: the inapplicable action and the next state.
        """
        self.logger.debug("Selecting an inapplicable action.")
        action_queue = list(ground_actions)
        random.shuffle(action_queue)
        inapplicable_action = action_queue.pop()
        operator = Operator(
            action=domain.actions[inapplicable_action.name],
            domain=domain,
            grounded_action_call=inapplicable_action.parameters,
            problem_objects=problem.objects,
        )
        while operator.is_applicable(current_state) and len(action_queue) > 0:
            inapplicable_action = action_queue.pop()
            operator = Operator(
                action=domain.actions[inapplicable_action.name],
                domain=domain,
                grounded_action_call=inapplicable_action.parameters,
                problem_objects=problem.objects,
            )

        return operator, State(predicates=current_state.state_predicates, fluents=current_state.state_fluents, is_init=False)

    def _select_applicable_action(
        self, domain: Domain, problem: Problem, current_state: State, grounded_actions: Set[ActionCall]
    ) -> Tuple[Operator, State]:
        """Selects a random applicable action for the current state.

        :param domain: the domain to select the applicable action from.
        :param problem: the problem defining the objects and the initial state.
        :param current_state: the current state to select the applicable action for.
        :param grounded_actions: the grounded actions to select the applicable action from.
        :return: the applicable action and the next state.
        """
        self.logger.debug("Selecting an applicable action.")
        action_queue = list(grounded_actions)
        random.shuffle(action_queue)
        action = action_queue.pop()
        operator = Operator(
            action=domain.actions[action.name], domain=domain, grounded_action_call=action.parameters, problem_objects=problem.objects
        )
        while not operator.is_applicable(current_state) and len(action_queue) > 0:
            self.logger.debug(f"Action {action} is inapplicable.")
            action = action_queue.pop()
            operator = Operator(
                action=domain.actions[action.name], domain=domain, grounded_action_call=action.parameters, problem_objects=problem.objects
            )

        if len(action_queue) == 0 and not operator.is_applicable(current_state):
            raise ValueError("No applicable actions found.")

        next_state = operator.apply(current_state, skip_validation=True)
        return operator, next_state

    def create_random_plan(self, domain: Domain, problem: Problem, grounded_actions: Set[ActionCall]) -> Tuple[List[TrajectoryTriplet], List[str]]:
        """Creates a random trajectory from the input domain and problem by randomly executing actions.

        :param domain: The domain that contain the actions to execute.
        :param problem: the problem containing the initial state and the objects.
        :param grounded_actions: a list of all the possible grounded actions that match the problem.
        :return: a list of the triplets that represent the trajectory and the plan.
        """
        self.logger.info(f"Starting to create a random trajectory for the problem - {problem.name}")
        plan = []
        current_state = State(predicates=problem.initial_state_predicates, fluents=problem.initial_state_fluents, is_init=True)
        action_triplets = []
        for i in range(MAX_NUM_STEPS_IN_TRAJECTORY):
            self.logger.debug(f"Selecting an action for timestep {i}.")
            if random.random() < inapplicable_action_probability:
                self.logger.debug(f"Selecting an inapplicable action for timestep {i}.")
                action, next_state = self._select_inapplicable_action(domain, problem, current_state, grounded_actions)
                action_triplets.append(TrajectoryTriplet(previous_state=current_state, op=action, next_state=next_state))
                plan.append(str(action))
                current_state = next_state.copy()
                continue

            self.logger.debug(f"Selecting an applicable action for timestep {i}.")
            try:
                action, next_state = self._select_applicable_action(domain, problem, current_state, grounded_actions)
                action_triplets.append(TrajectoryTriplet(previous_state=current_state, op=action, next_state=next_state))
                plan.append(str(action))
                current_state = next_state.copy()

            except ValueError:
                self.logger.warning(f"No applicable actions found for timestep {i}.")
                break

        return action_triplets, plan

    def create_all_grounded_actions(
        self, observed_objects: Dict[str, PDDLObject], domain: Domain, initial_state_fluents: List[str]
    ) -> Set[ActionCall]:
        """Creates all the grounded actions for the domain given the current possible objects.

        :param observed_objects: the objects that the learner has observed so far.
        :param domain: the domain of which the actions were taken from.
        :param initial_state_fluents: the initial state fluents to validate that all random actions do not interact with illegal fluents.
        :return: a set of all the possible grounded actions.
        """
        self.logger.info("Creating all the grounded actions for the domain given the current possible objects.")
        grounded_action_calls = self.vocabulary_creator.create_grounded_actions_vocabulary(domain=domain, observed_objects=observed_objects)
        legal_actions = set()

        for action in grounded_action_calls:
            operator = Operator(
                action=domain.actions[action.name], domain=domain, grounded_action_call=action.parameters, problem_objects=observed_objects
            )
            operator.ground()
            # validating that the preconditions do not contain any numeric fluents that do not exist in the problem.
            grounded_effect_fluents = set()
            for effect in operator.grounded_effects:
                grounded_effect_fluents.update(effect.grounded_numeric_fluents)

            if not (
                operator.grounded_preconditions.grounded_numeric_fluents.issubset(set(initial_state_fluents))
                and grounded_effect_fluents.issubset(set(initial_state_fluents))
            ):
                continue

            legal_actions.add(action)

        return legal_actions

    def create_domain_trajectories(self, problems_prefix: str = "pfile", output_directory_path: Optional[Path] = None) -> None:
        """Creates the domain trajectory files.

        :param problems_prefix: the prefix of the problem files.
        :param output_directory_path: the path to the output directory.
        """
        domain_file_path = self.working_directory_path / self.domain_file_name
        output_dir = output_directory_path or self.working_directory_path
        domain = DomainParser(domain_file_path).parse_domain()
        trajectory_exporter = TrajectoryExporter(domain=domain, allow_invalid_actions=False)
        for problem_file_path in self.working_directory_path.glob(f"{problems_prefix}*.pddl"):
            self.logger.info(f"Creating the trajectory for the problem - {problem_file_path.stem}")
            output_trajectory_path = output_dir / f"{problem_file_path.stem}_random_walk.trajectory"
            if output_trajectory_path.exists():
                continue

            problem = ProblemParser(problem_path=problem_file_path, domain=domain).parse_problem()
            grounded_actions = self.create_all_grounded_actions(
                observed_objects=problem.objects, domain=domain, initial_state_fluents=list(problem.initial_state_fluents.keys())
            )
            random_walk_triplets, plan = self.create_random_plan(domain, problem, grounded_actions)
            self.logger.debug("Creating a copy of the problem file with the trajectory as well as a solution_file.")
            with open(output_dir / f"{problem_file_path.stem}_random_walk.solution", "wt") as plan_file:
                plan_file.write("\n".join(plan))

            shutil.copy(problem_file_path, output_dir / f"{problem_file_path.stem}_random_walk.pddl")
            trajectory_exporter.export_to_file(random_walk_triplets, output_dir / f"{problem_file_path.stem}_random_walk.trajectory")


if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s %(name)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
    trajectory_creator = RandomWalkTrajectoriesCreator(domain_file_name=sys.argv[1], working_directory_path=Path(sys.argv[2]))
    selected_solver = SolverType.enhsp
    trajectory_creator.create_domain_trajectories()



"""This module contains the ConvexHullLearner class."""

from typing import Dict, List, Tuple, Optional, Union

import warnings
import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from pddl_plus_parser.models import Precondition, PDDLFunction
from scipy.spatial import ConvexHull, QhullError

from sam_learning.core.exceptions import NotSafeActionError
from sam_learning.core.learning_types import ConditionType, EquationSolutionType
from sam_learning.core.numeric_learning.convex_hull_learner import ConvexHullLearner
from sam_learning.core.numeric_learning.numeric_utils import (
    prettify_coefficients,
    construct_numeric_conditions,
    construct_projected_variable_strings,
    extended_gram_schmidt,
    display_convex_hull,
    create_monomials,
    create_polynomial_string,
    divide_span_by_common_denominator,
    remove_complex_linear_dependencies,
    EPSILON,
    construct_pddl_inequality_scheme,
)

warnings.simplefilter(action="ignore", category=FutureWarning)


class IncrementalConvexHullLearner(ConvexHullLearner):
    """Class that learns the convex hull for the preconditions of an action - incrementally."""

    _convex_hull: Optional[ConvexHull]
    _gsp_base: Optional[List[List[float]]]
    _complementary_base: Optional[List[List[float]]]
    _spanning_standard_base: bool
    data: DataFrame
    relevant_fluents: Optional[List[str]]
    affine_independent_data: DataFrame
    additional_dependency_conditions: List[str]

    def __init__(
        self,
        action_name: str,
        domain_functions: Dict[str, PDDLFunction],
        polynom_degree: int = 0,
        epsilon: float = 0.0,
        qhull_options: str = "",
        relevant_fluents: Optional[List[str]] = None,
    ):
        super().__init__(
            action_name,
            {function.name: function for function in domain_functions.values()},
            polynom_degree=polynom_degree,
        )
        self._convex_hull = None
        self._gsp_base = None
        self._complementary_base = None
        self._spanning_standard_base = False
        self.polynom_degree = polynom_degree
        functions = list([function.untyped_representation for function in domain_functions.values()])
        self.monomials = create_monomials(functions, polynom_degree)
        self.data = DataFrame(columns=[create_polynomial_string(monomial) for monomial in self.monomials])
        self.relevant_fluents = relevant_fluents if relevant_fluents is not None else self.data.columns.tolist()
        self.affine_independent_data = None
        self.additional_dependency_conditions = []
        self._epsilon = epsilon
        self._qhull_options = qhull_options

    @staticmethod
    def _calculate_orthonormal_complementary_base(
        orthonormal_base: List[List[float]], num_dimensions: int
    ) -> List[List[float]]:
        """Calculates the orthonormal complementary base of the given orthonormal base.

        :param orthonormal_base: the orthonormal base spanning the points in the storage dataframe.
        :param num_dimensions: the number of dimensions of the original space.
        :return: the orthonormal complementary base of the given orthonormal base.
        """
        diagonal_eye = [list(vector) for vector in np.eye(num_dimensions)]
        return extended_gram_schmidt(diagonal_eye, orthonormal_base)

    @staticmethod
    def _calculate_orthonormal_base(points_df: DataFrame) -> List[List[float]]:
        """Calculates the orthonormal base of the given points.

        :param points_df: the points to calculate the orthonormal base from.
        :return: the orthonormal base of the given points.
        """
        points = points_df.to_numpy()
        shift_axis = points[0].tolist()  # selected the first vector to be the start of the axis.
        shifted_points = points - shift_axis
        return extended_gram_schmidt(shifted_points)

    def _shift_new_point(self, point: Union[Series, DataFrame]) -> np.ndarray:
        """Shifts the points based on the first sample in the dataframe.

        Note: if the points are spanning the original space, we do not need to shift the points.

        :return: the shifted point based on the first sample in the dataframe.
        """
        numpy_sample = point.to_numpy()
        if (isinstance(point, DataFrame) and point.shape[1]) == 1 or (
            isinstance(point, Series) and point.shape[0] == 1
        ):
            self.logger.debug("The point is a single dimensional.")
            return numpy_sample - self.affine_independent_data.iloc[0][0]

        shifted_sample = (
            numpy_sample - self.affine_independent_data.iloc[0] if not self._spanning_standard_base else numpy_sample
        )
        return shifted_sample

    def _is_spanned_in_base(self, point: Series) -> bool:
        """Checks if the new point is spanned by the base.

        :return: whether the new point is spanned by the base.
        """
        shifted_sample = self._shift_new_point(point)
        return len(extended_gram_schmidt([list(shifted_sample)], self._gsp_base)) == 0

    def _learn_new_bases(self) -> None:
        """Learns the new bases of the points in the storage dataframe.

        :return: the new bases of the points in the storage dataframe.
        """
        self.logger.debug("Learning the new normal and the complementary bases of the points in the storage dataframe.")
        self._gsp_base = self._calculate_orthonormal_base(self.affine_independent_data)
        self._complementary_base = self._calculate_orthonormal_complementary_base(
            self._gsp_base, self.affine_independent_data.shape[1]
        )
        self._complementary_base = divide_span_by_common_denominator(self._complementary_base)
        if (len(self._gsp_base) > 1) and (len(self._gsp_base) == self.affine_independent_data.shape[1]):
            self.logger.debug("The points are spanning the original space, no need to project the points.")
            self._gsp_base = [list(vector) for vector in np.eye(self.affine_independent_data.shape[1])]
            self._spanning_standard_base = True

    def _learn_new_convex_hull(self, incremental: bool = True) -> None:
        """Learns the convex hull of the points in the storage dataframe.

        :param incremental: whether to use the incremental convex hull algorithm.
        :return: the convex hull of the points in the storage dataframe.
        """
        if self._convex_hull is not None:
            self._convex_hull.close()
            self._convex_hull = None

        self.logger.debug("Creating the convex hull for the first time (or in case that the base had changed).")
        points = self.affine_independent_data.to_numpy()
        shift_axis = (
            points[0].tolist()
            if not self._spanning_standard_base
            else [0] * len(self.affine_independent_data.columns.tolist())
        )
        projected_points = np.dot(points - shift_axis, np.array(self._gsp_base).T)
        self._convex_hull = self._epsilon_approximate_hull(points=projected_points, incremental=incremental)

    def _calculate_basis_and_hull(self, from_reset: bool = False) -> None:
        """Calculates the basis and the convex hull of the points in the storage dataframe."""
        prev_conditions = self.additional_dependency_conditions
        # Avoid cases when some of the data points are NaN and thus removed.
        self.affine_independent_data, self.additional_dependency_conditions = remove_complex_linear_dependencies(
            self.data[[col for col in self.data.columns if col in self.relevant_fluents]]
        )
        if self.affine_independent_data.empty:
            self.logger.debug("There are no affine independent rows, no need for further processing.")
            return

        # If dependency conditions changed, retrain base and CH (if needed)
        if prev_conditions != self.additional_dependency_conditions or from_reset or self._gsp_base is None:
            self.logger.debug(
                "Either new point affects the dependency conditions or no base learned. Learning base and possibly CH."
            )
            self._learn_new_bases()
            if self._gsp_base is not None and len(self._gsp_base) >= 2 and len(self.data) >= 3:
                self._learn_new_convex_hull()

            return

        # Validate span and try to add to convex hull
        if not self._is_spanned_in_base(self.affine_independent_data.iloc[-1]):
            self.logger.debug("New point is not spanned by base. Relearning base and CH.")
            self._learn_new_bases()
            if len(self._gsp_base) <= 1:
                self.logger.debug("New base is 1D or empty. Cannot form a convex hull.")
                return

            self._learn_new_convex_hull()
            return

        if self._convex_hull is not None:
            # If reached here - the point is spanned by the base and the convex hull is not None
            projected_point = np.dot(
                self._shift_new_point(self.affine_independent_data.iloc[-1]), np.array(self._gsp_base).T
            )
            self._convex_hull.add_points([projected_point])

    def add_new_point(self, point: Dict[str, float]) -> None:
        """Adds a new point to the convex hull learner.

        Note:
            This method is supposed to improve the performance of the CH calculations by incrementally adding points.

        :param point: the point to add to the convex hull learner.
        """
        new_sample = pd.DataFrame({k: [v] for k, v in point.items()})
        # check if the new sample is already in the dataframe
        concat_data = pd.concat([self.data, new_sample], ignore_index=True).dropna(axis=1)
        if concat_data.duplicated().any():
            self.logger.debug("The new point is already in the storage, not adding it again.")
            return

        self.data = concat_data.dropna(axis=1)
        self._calculate_basis_and_hull()

    def _create_ch_coefficients_data(self, display_mode: bool = True) -> Tuple[List[List[float]], List[float]]:
        """Runs the convex hull algorithm on the given input points.

        :param display_mode: whether to display the convex hull.
        :return: the coefficients of the planes that represent the convex hull and the border point.
        """
        display_convex_hull(self.action_name, display_mode, self._convex_hull)
        equations = np.unique(self._convex_hull.equations, axis=0)
        A = equations[:, : self._convex_hull.points.shape[1]]
        b = -equations[:, self._convex_hull.points.shape[1]]
        coefficients = [prettify_coefficients(row) for row in A]
        border_point = prettify_coefficients(b)
        return coefficients, border_point

    def _incremental_create_ch_inequalities(
        self, display_mode: bool = False
    ) -> Tuple[List[List[float]], List[float], List[str], Optional[List[str]]]:
        """Create the convex hull and returns the matrix representing the inequalities.

        :return: the matrix representing the inequalities of the planes created by the convex hull as well as the
            names of the features that are part of the convex hull.

        Note: the returned values represents the linear inequalities of the convex hull, i.e.,  Ax <= b.
        """
        if self._gsp_base is None:
            self.logger.debug("The base is not yet learned since didn't receive enough points.")
            raise ValueError()

        points = self.affine_independent_data.to_numpy()
        shift_axis = points[0].tolist()  # selected the first vector to be the start of the axis.

        if (len(self._gsp_base) > 1) and (len(self._gsp_base) == self.affine_independent_data.shape[1]):
            self.logger.debug("The points are spanning the original space and the basis is full rank.")
            coefficients, border_point = self._create_ch_coefficients_data(display_mode)
            return coefficients, border_point, self.affine_independent_data.columns.tolist(), None

        projected_points = np.dot(points - shift_axis, np.array(self._gsp_base).T)

        if self._convex_hull is None:
            self.logger.debug("The projected points are 1D, creating min-max conditions on the new base.")
            coefficients = [[-1], [1]]
            border_point = prettify_coefficients([-projected_points.min(), projected_points.max()])

        else:
            coefficients, border_point = self._create_ch_coefficients_data(display_mode)

        transformed_vars = construct_projected_variable_strings(
            self.affine_independent_data.columns.tolist(), shift_axis, self._gsp_base
        )
        self.logger.debug("Constructing the conditions to verify that points are in the correct span.")
        transformed_orthonormal_vars = construct_projected_variable_strings(
            self.affine_independent_data.columns.tolist(),
            shift_axis,
            [list(vector) for vector in np.eye(points.shape[1])],
        )
        span_verification_conditions = construct_pddl_inequality_scheme(
            np.array(self._complementary_base),
            np.zeros(len(self._complementary_base)),
            transformed_orthonormal_vars,
            sign_to_use="=",
        )
        return coefficients, border_point, transformed_vars, span_verification_conditions

    def construct_safe_linear_inequalities(self, *args, **kwargs) -> Precondition:
        """Constructs the linear inequalities strings that will be used in the learned model later.

        :return: the inequality strings and the type of equations that were constructed (injunctive / disjunctive)
        """
        if len(self.data) == 0 or len(self.relevant_fluents) == 0:
            self.logger.debug(
                "No observations were given or the relevant fluents are empty - no conditions could be learned."
            )
            return Precondition("and")

        if len(self.data.columns.tolist()) == 1 or len(self.relevant_fluents) == 1:
            self.logger.debug("Only one dimension is needed in the preconditions!")
            return self._construct_single_dimension_inequalities(self.data.loc[:, self.data.columns.tolist()[0]])

        try:
            if len(self.affine_independent_data) == 0:
                self.logger.warning("The matrix is empty, no need to create a convex hull.")
                return construct_numeric_conditions(
                    self.additional_dependency_conditions,
                    condition_type=ConditionType.conjunctive,
                    domain_functions=self.domain_functions,
                )

            A, b, column_names, additional_projection_conditions = self._incremental_create_ch_inequalities()
            self.logger.debug(f"Constructing the PDDL inequality scheme for the action {self.action_name}.")
            inequalities_strs = construct_pddl_inequality_scheme(A, b, column_names)
            if additional_projection_conditions is not None:
                inequalities_strs.extend(additional_projection_conditions)

            self.logger.debug("Constructing the precondition object from the constructed strings.")
            return construct_numeric_conditions(
                [*inequalities_strs, *self.additional_dependency_conditions],
                condition_type=ConditionType.conjunctive,
                domain_functions=self.domain_functions,
            )

        except (QhullError, ValueError):
            self.logger.warning(
                "Convex hull failed to create a convex hull, using disjunctive preconditions "
                "(probably since the rank of the matrix is 2 and it cannot create a degraded "
                "convex hull)."
            )
            raise NotSafeActionError(
                name=self.action_name,
                reason="Convex hull failed to create a convex hull",
                solution_type=EquationSolutionType.convex_hull_not_found,
            )

    def is_point_in_convex_hull(self, point: Union[Series, DataFrame]) -> bool:
        """Checks if the given point is in the convex hull.

        :param point: the point to check.
        :return: whether the point is in the convex hull.
        """
        concat_data = (
            pd.concat([self.data, point], ignore_index=True)
            if isinstance(point, DataFrame)
            else pd.concat([self.data, point.to_frame().T], ignore_index=True)
        )
        if concat_data.duplicated().any():
            self.logger.debug("The new point is already in the storage, it is thus is the convex hull.")
            return True

        if self._gsp_base is None:
            self.logger.debug("The convex hull is not yet learned since didn't receive enough points.")
            return False

        new_entry = point[self.affine_independent_data.columns.tolist()]
        if not self._is_spanned_in_base(new_entry):
            self.logger.debug("The point is not spanned by the base.")
            return False

        shifted_new_entry = self._shift_new_point(new_entry)
        projected_entry = np.dot(shifted_new_entry, np.array(self._gsp_base).T)
        if (
            len(self._complementary_base) > 0
            and (np.absolute(np.dot(np.array(self._complementary_base), np.array(shifted_new_entry).T)) > EPSILON).any()
        ):
            self.logger.debug("The new points are not in the span of the input points.")
            return False

        if projected_entry.shape[0] > 1:
            self.logger.debug("Validating whether the new point conforms with the CH inequalities.")
            A = self._convex_hull.equations[:, :-1]
            b = self._convex_hull.equations[:, -1]
            return bool(np.all(A.dot(projected_entry) + b <= EPSILON))

        self.logger.debug("The new point is single dimensional... checking if it is within the legal interval.")
        projected_dataset = np.dot(
            self.affine_independent_data.to_numpy() - self.affine_independent_data.to_numpy()[0],
            np.array(self._gsp_base).T,
        )
        return projected_dataset.min() <= projected_entry <= projected_dataset.max()

    def reset(self, new_relevant_features: List[str]) -> None:
        """Resets changes the features used to create the convex hull and reset to start the learning from scratch.

        :param new_relevant_features: the new relevant features to use.
        """
        self.logger.debug("Resetting the convex hull learner.")
        self.relevant_fluents = new_relevant_features
        self._convex_hull.close()
        self._convex_hull = None
        self._calculate_basis_and_hull(from_reset=True)

    def copy(self, one_shot: bool = False) -> "IncrementalConvexHullLearner":
        """Creates a copy of the current object.

        :param one_shot: whether to create a one-shot copy of the covex hull.
        :return: a copy of the current object.
        """
        new_learner = IncrementalConvexHullLearner(
            action_name=self.action_name,
            domain_functions=self.domain_functions,
            polynom_degree=self.polynom_degree,
            epsilon=self._epsilon,
            qhull_options=self._qhull_options,
            relevant_fluents=self.relevant_fluents,
        )
        if self.data.empty:
            self.logger.debug("The data is empty, no need to copy the convex hull.")
            return new_learner

        new_learner.data = self.data.copy()
        new_learner.affine_independent_data = self.affine_independent_data.copy()
        new_learner.additional_dependency_conditions = self.additional_dependency_conditions.copy()
        new_learner._gsp_base = self._gsp_base.copy() if self._gsp_base is not None else None
        new_learner._complementary_base = (
            self._complementary_base.copy() if self._complementary_base is not None else None
        )
        new_learner._spanning_standard_base = self._spanning_standard_base
        if self._convex_hull is not None:
            new_learner._learn_new_convex_hull(incremental=not one_shot)

        return new_learner

    def close_convex_hull(self) -> None:
        """Closes the convex hull."""
        if self._convex_hull is not None:
            self._convex_hull.close()
            self._convex_hull = None

    def is_spanning_standard_base(self) -> bool:
        """Checks if the convex hull is spanning the standard base.

        :return: whether the convex hull is spanning the standard base.
        """
        return self._spanning_standard_base
